functor ProofRulesAthenaFun(structure Hash: HASH
			    structure Interface: INTERFACE): 
  PROOF_RULES_ATHENA =

  struct

    structure Sequent = 
      SequentAthenaFun(structure Hash = Hash
			structure Interface = Interface)

    open Sequent
    open InputLanguage
    open Typecheck

    open Hash
    open Interface
    open Options

    open SympBug
    open Str
    open Pos
    open ListOps
    open Hash
    open ParserAthena
    open ParseTreeStruct

    exception NoGoals

    type ProofSystemTactic = unit
    type ProofSystemTacticState = unit

    (* Each rule has a unique string identifier *)
    fun getRuleName ({name=name,...}:InferenceRule) = name
    fun getCommandName ({name=name,...}:ProofSystemCommand) = name

    (* Each rule has a set of arguments *)
    fun getRuleArgs ({params=params,...}:InferenceRule) = params
    fun getCommandArgs ({params=params,...}:ProofSystemCommand) = params

    fun getRuleHelp ({help=help,
		      name=name,
		      params=params, ...}:InferenceRule) =
	(formatComm(name, List.map InfRuleSpec2str params)^"\n\n"^help)

    fun getCommandHelp ({help=help,
			 name=name,
			 params=params, ...}:ProofSystemCommand) =
	(formatComm(name, List.map InfRuleSpec2str params)^"\n\n"^help)

    fun getTacticName tactic = "<tactic name>"
    fun getTacticArgs tactic = []: InferenceRuleArgumentSpec list
    fun getTacticHelp tactic = "<tactic help>"
      
    (* Functions for finding goals in the initial nodes. *)

    fun isName(Id _) = true
      | isName(TypedId _) = true
      | isName(UniqueId _) = true
      | isName _ = false

    fun allGoals n (Labeled (_, p)) = allGoals n p
      | allGoals n (Send (_, _)) = []
      | allGoals n (Receive (_, m)) = allGoals n m
      | allGoals n (MessageTuple (_, mlist)) =
	let val mresults = (List.map (allGoals n) mlist)
	    fun doOne(i, goals) = List.map (fn Goal(r, pt) => Goal(ExtractTuple (r, i), pt)) goals
	in
	    List.foldr op@ [] (mapEnumerate doOne mresults)
	end
      | allGoals n (MessageConcur (_, mlist)) =
	let val mresults = (List.map (allGoals n) mlist)
	    fun doOne(i, goals) = List.map (fn Goal(r, pt) => Goal(ExtractConcur (r, i), pt)) goals
	in
	    List.foldr op@ [] (mapEnumerate doOne mresults)
	end
      (* Filter goals that ask for publicly known info, namely 
         pub. keys and names of principals. *)
      | allGoals n (x as Pub(_, k)) = if isName k then [] else [Goal(ExtractMessage n, x)]
      | allGoals n x =
	 let val tp = messageType x
	 in
	     case tp of
		 PrincipalType _ => if isName x then [] else [Goal(ExtractMessage n, x)]
	       | _ => [Goal(ExtractMessage n, x)]
	 end

    fun findGoals(node, _) = allGoals node (nodeAction node)

    (* Generate the set of unbound goals in the given set of nodes,
       and add them to the optional set of goals, or create a new set. *)
    fun generateGoals nodeHash =
	let val allNodes = hash2any (fn x=>x) nodeHash
	in 
	    List.foldr(op @) [] (List.map findGoals allNodes)
	end

    (* Functions that parse through roles for insertion into hash tables. *)

    fun findInBody (label, [], _) =
      raise SympBug (label ^ ": unmatched label not detected by typechecker!")
      | findInBody (label, Labeled (UniqueId (_, uid, _), _)::t, n) =
      if (label = uid) then 
	n
      else
	findInBody (label, t, n + 1)
      | findInBody (label, h::t, n) =
      findInBody (label, t, n + 1)

    fun findLabel (label, Role (_,{body=RoleBody(_,b),...})) =
      findInBody (label, b, 1)
      | findLabel (_, exp) =
      raise SympBug ((pt2string exp) ^ 
		     ": unexpected non Role/RoleBody in findLabel!")

    fun extractBody (Role (_, {body = RoleBody (_, b),...})) =
      b
      | extractBody _ = raise SympBug 
      "Attempted to extract body from non-Role!"

    (* Function that adds roles and nodes into a particular state
       sequent's hash tables. *)

    fun addNodesAndRoles context (roleHash, nodeHash: (Node, unit) Hash)
      (r as RoleLabel (p1, UniqueId(p2,rid,_), sparameters, 
		       UniqueId(p3,label,_))) =
      let
	val (role, pa) = extractRoleParams
	  (case 
	     (List.find 
	      (fn (Role (_, {name=UniqueId(_,x,_),...})) => 
	       x = rid |
	       _ => false) context) of
	     SOME r => r
	   | NONE => 
	       raise SympBug 
		 ("Typechecker failed to detect undefined role reference:\n  "
		  ^(pt2stringVerbose r)))
	val role' = instantiate (role, pa, sparameters)
	val b = extractBody role'
	val roleLine = findLabel (label, role)
	fun forLoop nid x =
	    (insertHashDestructive (nodeHash, 
				    makeNode(nid, x, rid, sparameters, List.nth (b, x-1)),
				    ());
	     if x = roleLine then () else forLoop nid (x + 1))
      in
	let
	  val allRoles = hash2any (fn x=>x) roleHash
	in
	  (case List.find (fn (x, y) => ptEq (y, role')) allRoles of
	     SOME (nid', _) =>
	       (forLoop nid' 1;
		())
	   | NONE =>
		  let val nid = genSym rid
		  in 
		    (insertHashDestructive (roleHash, nid, role');
		     forLoop nid 1;
		     ())
		  end)
	end
      end
      | addNodesAndRoles context (roleHash, nodeHash)
      (r as RoleLabel (p1, UniqueId(p2,rid,_), sparameters, 
		       Range(p3,UniqueId(p4,label1,_),
			     UniqueId(p5,label2,_)))) =
      let
	val (role, pa) = extractRoleParams
	  (case 
	     (List.find 
	      (fn (Role (_, {name=UniqueId(_,x,_),...})) => 
	       x = rid |
	       _ => false) context) of
	     SOME r => r
	   | NONE => 
	       raise SympBug 
		 ("Typechecker failed to detect undefined role reference:\n  "
		  ^(pt2stringVerbose r)))
	val role' = instantiate (role, pa, sparameters)
	val b = extractBody role'
	val roleLine1 = findLabel (label1, role)
	val roleLine2 = findLabel (label2, role)
	fun forLoop nid x =
	  if x = roleLine2 then
	    (insertHashDestructive (nodeHash,
				    makeNode(nid, x, rid, sparameters, List.nth (b,x-1)),
				    ());
	     ())
	  else
	    (insertHashDestructive (nodeHash,
				    makeNode(nid, x, rid, sparameters, List.nth (b,x-1)),
				    ());
	     forLoop nid (x + 1))
      in
	let
	  val allRoles = hash2any (fn x=>x) roleHash
	in
	  (case List.find (fn (x, y) => ptEq (y, role')) allRoles of
	     SOME (nid', _) =>
	       (forLoop nid' 1;
		())
	   | NONE =>
	       let val nid = genSym rid
	       in
		 (insertHashDestructive (roleHash, nid, role');
		  forLoop nid 1;
		  ())
	       end)
	end
      end
      | addNodesAndRoles context (roleHash, nodeHash)
      (r as RoleStrand (p, UniqueId(_,rid,_), sparameters)) =
            let
	val (role, pa) = extractRoleParams
	  (case 
	     (List.find 
	      (fn (Role (_, {name=UniqueId(_,x,_),...})) => 
	       x = rid |
	       _ => false) context) of
	     SOME r => r
	   | NONE => 
	       raise SympBug 
		 ("Typechecker failed to detect undefined role reference:\n  "
		  ^(pt2stringVerbose r)))
	val role' = instantiate (role, pa, sparameters)
	val b = extractBody role'
	val length = List.length b
	fun forLoop nid x =
	  if (x = length) then
	    (insertHashDestructive (nodeHash,
				    makeNode(nid, x, rid, sparameters, List.nth (b, x-1)),
				    ());
	     ())
	  else
	    (insertHashDestructive (nodeHash,
				    makeNode(nid, x, rid, sparameters, List.nth (b, x-1)),
				    ());
	     forLoop nid (x + 1))
      in
	   let
	     val allRoles = hash2any (fn x=>x) roleHash
	   in
	     (case List.find (fn (x, y) => ptEq (y, role')) allRoles of
		SOME (nid', _) =>
		  (forLoop nid' 1;
		   ())
	      | NONE =>
		  let val nid = genSym rid
		  in (insertHashDestructive (roleHash, nid, role');
		      forLoop nid 1;
		      ())
		  end)
	   end
      end
    | addNodesAndRoles context (roleHash, nodeHash) exp =
      raise SympBug ((pt2string exp) ^ 
		     ": this should not happen (addNodesAndRoles)!")

    fun makeBundle roleHash (strandList, context) =
      let
	val nodeHash = makeHashDefault (nodeEq, nodeString)
      in
	(List.app
	 (fn Atomic (p, slist) => 
	  List.app (addNodesAndRoles context (roleHash, nodeHash)) slist |
	  exp => raise SympBug 
	    ((pt2string exp) ^ ": should not be non-atomic! (makeBundle)"))
	 strandList;
	 (roleHash, nodeHash))
      end

    (* Various functions for driving negations, etc. and transforming 
       formula into CNF form for state sequents. *)

    fun driveNegs (exp as FormulaInstance (p, UniqueId(p1,uid,id), sparams), 
		  context) =
      (case 
	 (List.find 
	  (fn (PredicateContext (_,{name=UniqueId(_,x,_),...},_)) => (x=uid) |
	   (TheoremContext (_, {name=UniqueId(_,x,_), ...}, _)) => (x=uid) |
	     _ => false) context) of
	 SOME (PredicateContext (_, {def=d, params=pa,...}, innerContext)) =>
	   driveNegs (instantiate (d, pa, sparams), innerContext)
       | SOME (TheoremContext (_, {def=d,params=pa,...}, innerContext)) =>
	   driveNegs (instantiate (d, pa, sparams), innerContext)
       | SOME _ =>
	   raise SympBug "This case should not be possible (driveNegs 1)."
       | NONE => 
	   raise SympBug 
	     ((pt2string exp) ^ 
	      ": type checker produced reference without context!")) 
      | driveNegs (exp as RoleStrand (p, UniqueId(p1,uid,id), sparams), 
		  context) =
      (case 
	 (List.find 
	  (fn (PredicateContext (_,{name=UniqueId(_,x,_),...},_)) => (x=uid) |
	   (TheoremContext (_, {name=UniqueId(_,x,_), ...}, _)) => (x=uid) |
	     _ => false) context) of
	 SOME (PredicateContext (_, {def=d, params=pa,...}, innerContext)) =>
	   driveNegs (instantiate (d, pa, sparams), innerContext)
       | SOME (TheoremContext (_, {def=d,params=pa,...}, innerContext)) =>
	   driveNegs (instantiate (d, pa, sparams), innerContext)
       | SOME _ =>
	   raise SympBug "This case should not be possible (driveNegs 2)."
       | NONE => 
	   raise SympBug 
	     ((pt2string exp) ^ 
	      ": type checker produced reference without context!")) 
      | driveNegs (exp as Not (p, FormulaInstance 
			       (_, UniqueId(_,uid,id), sparams)), 
		   context) =
      (case 
	 (List.find 
	  (fn (PredicateContext (_,{name=UniqueId(_,x,_),...},_)) => (x=uid) |
	   (TheoremContext (_, {name=UniqueId(_,x,_), ...}, _)) => (x=uid) |
	     _ => false) context) of
	 SOME (PredicateContext (_, {def=d, params=pa,...}, innerContext)) =>
	   driveNegs (Not (p, instantiate (d, pa, sparams)), innerContext)
       | SOME (TheoremContext (_, {def=d,params=pa,...}, innerContext)) =>
	   driveNegs (Not (p, instantiate (d, pa, sparams)), innerContext)
       | SOME _ =>
	   raise SympBug "This case should not be possible (driveNegs 3)."
       | NONE => 
	   raise SympBug 
	     ((pt2string exp) ^ 
	      ": type checker produced reference without context!")) 
      | driveNegs (exp as Not (p, RoleStrand 
			       (_, UniqueId(_,uid,id), sparams)), 
		   context) =
      (case 
	 (List.find 
	  (fn (PredicateContext (_,{name=UniqueId(_,x,_),...},_)) => (x=uid) |
	   (TheoremContext (_, {name=UniqueId(_,x,_), ...}, _)) => (x=uid) |
	     _ => false) context) of
	 SOME (PredicateContext (_, {def=d, params=pa,...}, innerContext)) =>
	   driveNegs (Not (p, instantiate (d, pa, sparams)), innerContext)
       | SOME (TheoremContext (_, {def=d,params=pa,...}, innerContext)) =>
	   driveNegs (Not (p, instantiate (d, pa, sparams)), innerContext)
       | SOME _ =>
	   raise SympBug "This case should not be possible (driveNegs 4)."
       | NONE => 
	   raise SympBug 
	     ((pt2string exp) ^ 
	      ": type checker produced reference without context!")) 
      | driveNegs (Implies (p, f1, f2), context) =
	 Or (p, driveNegs (Not (p, f1), context), driveNegs (f2, context))
      | driveNegs (And (p, f1, f2), context) =
	 And (p, driveNegs (f1, context), driveNegs (f2, context))
      | driveNegs (Or (p, f1, f2), context) =
	 Or (p, driveNegs (f1, context), driveNegs (f2, context))
      | driveNegs (Not (p1, Not (p2, f)), context) =
	 driveNegs (f, context)
      | driveNegs (Not (p1, Implies (p2, f1, f2)), context) =
	 And (p1, driveNegs (f1, context), driveNegs (Not (p2, f2), context))
      | driveNegs (Not (p1, And (p2, f1, f2)), context) =
	 Or (p1, driveNegs (Not (p2, f1), context), driveNegs (Not (p2, f2), 
							       context))
      | driveNegs (Not (p1, Or (p2, f1, f2)), context) =
	 And (p1, driveNegs (Not (p2, f1), context), driveNegs (Not (p2, f2), 
								context))
      | driveNegs (f, context) = 
	 f


    fun dist (And (p, f1, f2), f3) = And (p, dist (f1, f3), dist (f2, f3))
      | dist (f1, And (p, f2, f3)) = And (p, dist (f1, f2), dist (f1, f3))
      | dist (f1, f2) = Or (pos f1, f1, f2)

    fun CNF (And (p, f1, f2)) = And (p, CNF f1, CNF f2)
      | CNF (Or (p, f1, f2)) = dist (CNF f1, CNF f2)
      | CNF f = f

    fun orList (Or (p, f1, f2)) = (orList f1) @ (orList f2)
      | orList f = [f]
      
    fun toImp l = 
	let
	    val (n, p) = List.partition (fn (Not x) => true | _ => false) l
	in
	    [(List.map (fn (Not (_, x)) => x | e => 
			raise SympBug 
			    (pos2string (pos e) ^ 
			     ": Expected a not (athena-proof-rules)!")) 
	      n, p)]
	end

    fun andList (And (p, f1, f2)) = (andList f1) @ (andList f2)
      | andList f = toImp (orList f)

    fun prepare (f, context) =
      let val funName = "ProofRules.prepare"
	  val debug = lazyVerbDebug(getOptions()) funName
	val negNorm = driveNegs (f, context)
	val f' = CNF negNorm
	val _ = debug(fn()=>funName^": CNF form = "^(pt2string f')^"\n")
      in
	andList f'
      end

    (* Functions that split delta into its strands so that the final
       rule can test for intersection. *)

    fun eachStrand [] = []
      | eachStrand (sl as ((n, ())::t)) =
      let val s = nodeStrand n
	val (strand, others) = List.partition (fn (n', ()) => (nodeStrand n') = s) sl
      in
	strand :: (eachStrand others)
      end

    fun intoStrands (_, nodeHash) =
      let
	val allNodes = hash2any (fn x=>x) nodeHash
      in
	eachStrand allNodes
      end

  (* Check if the list of nodes is all in the nodeHash.  We match the
     nodes by role, its parameters, and the index.  So, even if the
     strands can be different, they may be *semantically* the same,
     and this is what we are checking.

     This returns a function (Node list -> bool). *)

    fun subStrand nodeHash =
	let val funName = "subStrand"
	    val debug = lazyVerbDebug(getOptions()) funName
	    fun hash2str hash =
		let fun doOne((s, i), nodes) =
		          "("^s^"."^(Int.toString i)^": [\n"
			  ^(strlistmap2str ",\n" nodeStringDebug nodes)^"])"
		in
		    "["^(strlist2str ",\n" (hash2any doOne hash))^"]"
		end
	    (* First, rehash nodes by their role id and index *)
	    val hash = makeHashDefault(op =, fn(s, i)=> s^"."^(Int.toString i))
	    fun addNode(n, _) =
		let val key = (nodeRole n, nodeIndex n)
		    val other = (case findHash(hash, key) of
				     SOME lst => lst
				   | NONE => [])
		in (insertHashDestructive(hash, key, n::other); ())
		end
	    val _ = hash2any addNode nodeHash
	    val _ = debug(fn()=>funName^": hash = "^(hash2str hash)^"\n")
	    fun compareParams(NONE, NONE) = true
	      | compareParams(SOME l1, SOME l2) =
		  (case zipOpt(l1, l2) of
		       SOME pairs => List.all ptEq pairs
		     | NONE => false)
	      | compareParams _ = false
	    fun checkNode (n, _) =
		(case findHash(hash, (nodeRole n, nodeIndex n)) of
		     NONE => false
		   | SOME lst => List.exists(fn n'=>compareParams(nodeRoleParams n,
								  nodeRoleParams n')) lst)
	    (* This is the result function *)
	    fun checkStrand nodes = List.all checkNode nodes
	in
	    checkStrand
	end

    (* Calls the appropriate functions for the Athena "final"
       rule. *)

    fun finalApply printFun (StateSequent 
			     {l = {semibundle = (_, nodeHash),
				   goalBinding = goalBinding,
				   goals = g,
				   typeMap = map, ...},
			      delta = delta, context = context, ...},
			     _, _, _) =
      let
	val deltaStrands = intoStrands delta
      in
	if (List.exists (subStrand nodeHash) deltaStrands) then
	  InferenceRuleSuccess []
	else
	  InferenceRuleFailure
      end
      | finalApply printFun _ =
      InferenceRuleFailure

    (* Functions for finding goals in the initial nodes. *)

    fun allBindings (n: Node) (Labeled (_, p)) = allBindings n p
      | allBindings n (Receive (_, _)) = []
      | allBindings n (Send (_, m)) = allBindings n m
      | allBindings n (MessageTuple (_, mlist)) =
	let val mresults = List.map (allBindings n) mlist
	    fun doOne(i, bindings) =
		  List.map (fn Binding(r, pt) => Binding(ExtractTuple (r, i), pt)) bindings
	in
	    List.foldr op@ [] (mapEnumerate doOne mresults)
	    (* (List.tabulate (List.length mlist, 
			    fn i => List.map
			    (fn (r, pt) => (ExtractTuple (r, i), pt))
			    (List.nth (mresults, i)))) *)
	end
      | allBindings n (MessageConcur (_, mlist)) =
      let
	val mresults = (List.map (allBindings n) mlist)
	fun doOne(i, bindings) =
	      List.map (fn Binding(r, pt) => Binding(ExtractConcur (r, i), pt)) bindings
      in
	List.foldr op@ [] (mapEnumerate doOne mresults)
	(* (List.tabulate (List.length mlist, 
			fn i => List.map
			(fn (r, pt) => (ExtractConcur (r, i), pt))
			(List.nth (mresults, i)))) *)
      end
      | allBindings n x =
      [Binding(ExtractMessage n, x)]

    fun findBindings nodeHash =
      let
	val allNodes = hash2any (fn x=>x) nodeHash
      in
	List.foldr op@ [] (List.map (fn (n, _) => allBindings n (nodeAction n)) allNodes)
      end

    (* Prepares CNF form, breaks into separate sequents, and creates 
       initial states. *)

    fun makeInitialSequents
      (TheoremContext (p, {name=n, params=pa, def=d}, theoremContext), pp: PruneParams) =
      let
	val gammaDeltaList = prepare (d, theoremContext)
      in
	List.map
	 (fn (gamma, delta) =>
	  let val roleHash = makeHashDefault (op=, fn x=>x)
	      val sb = makeBundle roleHash (gamma, theoremContext)
	  in
	    StateSequent {l = {semibundle = sb,
			       goalBinding = makeHashDefault (goalEq,
							      goalString),
			       intermConstr = [],
			       typeMap = makeHashDefault (op=, varString),
			       goals = sortList goalLess (generateGoals (getNodes sb))},
			  pruneParams = pp,
			  delta = makeBundle roleHash (delta, theoremContext),
			  context = theoremContext}
	  end)
	 gammaDeltaList
      end |
    makeInitialSequents (exp, _) =
    raise SympBug ((pt2string exp) ^ 
		   ": is not a TheoremContext! (makeInitialSequents)")

    (* Calls the appropriate functions for the Athena "init" 
       rule. *)

    fun initApply printFun (InitSequent{formula=pt, pruneParams=pp}, pcontext, args, substOpt) =
	  InferenceRuleSuccess ((List.map (fn x => (x, NONE))) 
				 (makeInitialSequents(pt, pp)))
      | initApply printFun _ = InferenceRuleFailure


    (* Functions for Most General Unifier *)

    fun InvertKey(Pub(p, k)) = Priv(p, k)
      | InvertKey(Priv(p, k)) = Pub(p, k)
      | InvertKey(InvKey(p, k)) = k
      | InvertKey k =
	(case messageType k of
	     SymKeyType _ => k
	   | _ => InvKey(pos k, k))

    fun typesMatch (m1, m2) = isSubtype(messageType m1, messageType m2)

    (* Substitution type for the Split rule *)
    type Subst = (string, ParseTree) Hash

    fun makeSubstitution() = makeHashDefault (op =, fn x=>x): Subst

    fun subst2str (S: Subst) =
	let fun pair2str(v, pt) = "("^v^" -> "^(pt2string pt)^")"
	in "Substitution["^(strlist2str ", " (hash2any pair2str S))^"]"
	end

    fun getUid(pt as UniqueId(_,uid,_)) = uid
      | getUid(TypedId(_, uid,_,_)) = uid
      | getUid x = raise SympBug
	  ("ProofRules.getUid: not a valid Id: "^(pt2stringVerbose x))

    (* Do the substitution on a ParseTree *)
    fun substitute subHash pt =
	let fun recur pt = ptTransformOne sf pt
	    and sf pt =
		let val funName = "ProofRules.substitute"
		    val _ = pushFunStackLazy(funName, fn()=> pt2string pt)
		    val res = sf' pt
		    val _ = popFunStackLazy(funName, fn()=> pt2string res)
		in res
		end
	    and sf'(pt as UniqueId(_,uid,_)) = 
	          (case findHash(subHash, uid) of
		       SOME x => sf x
		     | NONE => pt)
	      | sf'(pt as TypedId(_, uid, _, _)) =
	          (case findHash(subHash, uid) of
		       SOME x => sf x
		     | NONE => pt)
	      (* Collapse some expressions *)
	      | sf'(InvKey(_, pt)) = InvertKey(sf pt)
	      (* | sf'(Decrypt(p1, Crypt(p2, pt, k1), k2)) =
		let val pt' = sf pt
		    val (k1', k2') = (sf k1, sf k2)
		in
		    if ptEq(k1', InvertKey k2') then pt'
		    else Decrypt(p1, Crypt(p2, pt', k1'), k2')
		end *)
	      | sf' pt = recur pt
	in sf pt
	end
    
    (* Compute the most general unifier for two messages, so that m1 \interm m2.
       The MGU in our case is (C, S), where S is a normal substitution, and
       C is a set of constraints of the form t \interm t' (list of pairs). *)

    fun MGU (C, S) (m1, m2) =
	let val funName = "MGU"
            fun C2str C =
		"["^(strlistmap2str ",\n "
		     (fn(t,t')=>"("^(pt2string t)^" << "^(pt2string t')^")") C)
		^"]"
	    val _ = pushFunStackLazy(funName, fn()=>(pt2string m1)^", "^(pt2string m2)
				     ^"\n C = "^(C2str C))
	    exception NoUnifier

	    val debug = lazyVerbDebug(getOptions()) funName
	    fun addSubst(S, v, t) =
		let val t' = substitute S t
		    (* See if v occurs in t' *)
		    fun findVar pt = ptEq(v, pt) orelse List.exists findVar (ptSubtrees pt)
		in
		    if findVar t' then raise NoUnifier
		    else (insertHashDestructive (S, getUid v, t'); ())
		end
	    
	    (* Check if inverting the key simplifies it *)
	    fun reducesInvKey k =
		let val funName = funName^"/reducesInvKey"
		    val _ = pushFunStackLazy(funName, fn()=>pt2string k)
		    val res = (case InvertKey k of
				   InvKey _ => false
				 | _ => true)
		    val _ = popFunStack(funName, if res then "true" else "false")
		in
		    res
		end

	    (* Try to unify a variable with a non-variable term `m'.
	       Take special care of inverted keys; namely the terms
	       (INV (k: Key)) and (v: SymKey) can be unified, but by
	       assigning k, and not v.  Idea: if it doesn't match
	       directly, check if either side can be simplified by
	       inverting the key, and if one does, simplify it and
	       recurse. *)

	    fun matchVar S (v, m) =
		if typesMatch(m, v) then 
		    (case (findHash (S, getUid v)) of 
			 NONE => addSubst(S, v, m)
		       | SOME m' => loop S (m', m))
		else
		    (case m of
			 InvKey(_, k) =>
			     if reducesInvKey v then loop S (InvertKey v, k)
			     else if reducesInvKey k then loop S (v, InvertKey k)
			     else raise NoUnifier
		       | _ => raise NoUnifier)
			     
	    (* MGU for the exact substitution (making m1 = m2) *)
	    and loop S (m1, m2) =
		let val funName = funName^"/loop"
		    val _ = pushFunStackLazy(funName, fn()=>(pt2string m1)^", "^(pt2string m2))
		in (loop' S (m1, m2); popFunStackLazy(funName, fn()=>"Returning"))
		end
	    and loop' S (Crypt(_, m1, k1), Crypt (_, m2, k2)) =
	          (loop S (m1, m2); loop S (k1, k2))
	      | loop' S (MessageTuple(_, ml1), MessageTuple (_, ml2)) =
		  (case zipOpt(ml1, ml2) of
		       SOME pairs => List.app (loop S) pairs
		     | NONE => raise NoUnifier)
	      (* Special case: one var is mapped to the other.  We map
	         the var with the larger type to the one with the
	         smaller, to keep the type of both smaller. *)
	      | loop' S (v1 as TypedId _, v2 as TypedId _) =
		  if ptEq(v1, v2) then ()
		  else if typesMatch(v1, v2) then addSubst(S, v2, v1)
		  else if typesMatch(v2, v1) then addSubst(S, v1, v2)
		  else raise NoUnifier
	      | loop' S (v as TypedId _, m) = matchVar S (v, m)
	      | loop' S (m, v as TypedId _) = matchVar S (v, m)
	      | loop' S (Pub(_, m1), Pub(_, m2)) = loop S (m1, m2)
	      | loop' S (Priv(_, m1), Priv(_, m2)) = loop S (m1, m2)
	      | loop' S (SymKey(_, l1), SymKey(_, l2)) =
		  (case zipOpt(l1,l2) of
		       SOME pairs => List.app (loop S) pairs
		     | NONE => raise NoUnifier)
	      | loop' S (InvKey(_, k1), v2) = loop S (k1, InvertKey v2)
	      | loop' S (v1, InvKey(_, k2)) = loop S (InvertKey v1, k2)
	      | loop' S (m1, m2) =
		  if ptEq(m1, m2) then ()
		  else raise NoUnifier
	    fun unify S (m1, m2) =
		(loop S (m1, m2); SOME S) handle NoUnifier => NONE
	    (* Use as much info in C as possible to update S;
	       Return a list of pairs (C, S) of new constraints and the substitution.
	       If C found to be contradictory, return the empty list.
	       The steps are:
	       * Eliminate tautologies
	       * Check if type of RHS is >= Tuple; if not, two expressions must be unified;
	       * Find cycles in C, unify all the terms in each cycle and remove them from C;
	       * Check that no proper subterm cycles are present;
	       * If RHS is a tuple, then split cases on how the LHS can match.
	         Collect all possible matches.
	       * After each step, normalize C (apply the current substitution, 
		 split tuples on the LHS);
	       * Repeat these steps until C doesn't change. *)
            fun reduce(C, S) =
		let val funName = funName^"/reduce"
		    val _ = pushFunStackLazy(funName, fn()=>(subst2str S)^"\n C = "^(C2str C))
		    val debug = lazyVerbDebug(getOptions()) funName
		    val S = copyHash S
		    val changed = ref false (* whether C has changed or not *)
		    fun parts(m as MessageTuple(_, lst)) =
			  union ptEq (List.map parts lst)
		      | parts m = [m]
		    (* Instantiate C with the current substitution and
		       if the LHS is not atomic, split it into several inequalities *)
		    fun flatten C =
			let fun splitLHS [] = []
			      | splitLHS((m1,m2)::C) =
			         (case parts m1 of
				      [] => raise SympBug
					  "ProofRules.MGU/reduce/splitLHS: this can't happen"
				    | [ _ ] => (m1, m2)::(splitLHS C)
				    (* More than 1 parts *)
				    | lst => (changed := true;
					      (List.map (fn t=> (t, m2)) lst)@(splitLHS C)))
			    fun substPair(m1, m2) =
				(substitute S m1, substitute S m2)
			    fun eqC(C1, C2) = 
				(case zipOpt(C1, C2) of
				     SOME pairs => List.all(fn((t1,t2),(m1,m2))=>
							    ptEq(t1,m1) andalso ptEq(t2,m2)) pairs
				   | NONE => false)
			    val C' = List.map substPair C
			in
			    (if eqC(C, C') then () else (changed := true);
			     splitLHS C')
			end
		    (* Filter tautologies from C *)
		    fun tautology [] = []
		      | tautology ((m1, m2)::C) =
			if subset ptEq (parts m1, parts m2) then
			    (changed := true; tautology C)
			else (m1, m2)::(tautology C)
		    (* Filter out pairs with non-tuple types on the RHS *)
		    fun primType((m1, m2)::C) =
			let val tp2 = messageType m2
			    fun addPair NONE = NONE
			      | addPair (SOME lst) = SOME((m1, m2)::lst)
			in
			    case tp2 of
				TupleType _ => addPair(primType C)
			      | MessageType _ => addPair(primType C)
			      | _ => (case unify S (m1, m2) of
					  SOME _ => (changed := true;
						     primType(flatten C))
					| NONE => NONE)
			end
		      | primType [] = SOME []
		    (* Find and unify away all the cycles in C *)
		    fun checkCycles(C as (t, _)::_) =
		       (let val visitedHash = makeHashDefault(ptEq, pt2string)
			    fun mark pt = insertHash(visitedHash, pt, ())
			    fun visited pt = isSome(findHash(visitedHash, pt))
			    fun image t = List.mapPartial(fn (t1, t2) =>
							  if ptEq(t, t1) then SOME t2
							  else NONE) C
			    fun inStack(stack, t) =
				let fun loop([], t, acc) = NONE
				      | loop(t'::stack, t, acc) =
				        (* Interm cycle *)
				        if ptEq(t, t') then SOME acc
					(* Subterm cycle: this one is `fatal' *)
					else if subterm(t, t') then raise NoUnifier
					else loop(stack, t, t'::acc)
				in
				    loop(stack, t, [t])
				end
			    (* Look for a cycle and return it immediately if found *)
			    fun dfs(stack, t) =
				let fun dfslist(stack, []) = NONE
				      | dfslist(stack, t::rest) =
				         (case dfs(stack, t) of
					      res as SOME _ => res
					    | NONE => dfslist(stack, rest))
				in
				    case inStack(stack, t) of
					SOME lst => SOME lst
				      | NONE => if visited t then NONE
						else (mark t; dfslist(t::stack, image t))
				end
			    fun pairNotEq (t1, t2) (m1,m2) =
				not(ptEq(m1,t1) andalso ptEq(m2, t2))
			    (* Remove the cycle from C.  The cycle starts and ends with `t'.
			       The stack runs "backwards": [t3,t2,t1] means
			       (t1,t2), (t2,t3) and (t3, t1) are in C. *)
			    fun removeCycle([], t, C) = C
			      | removeCycle([t1], t2, C) = List.filter (pairNotEq(t1, t2)) C
			      | removeCycle(t2::(rest as t1::_), t, C) =
				  removeCycle(rest, t, List.filter(pairNotEq(t1,t2)) C)
			    (* Unify all the pairs on the cycle *)
			    fun unifyCycle([], t) = SOME S
			      | unifyCycle([t1], t2) = unify S (t1, t2)
			      | unifyCycle(t2::(rest as t1::_), t) =
				  (case unify S (t1, t2) of
				       SOME _ => unifyCycle(rest, t)
				     | NONE => NONE)
			in
			    case dfs([], t) of
				SOME(c as (t'::_)) =>
				    (case unifyCycle(c, t') of
					 SOME _ => 
					     (changed := true;
					      (* Keep checking for cycles *)
					      checkCycles(flatten(removeCycle(c, t', C))))
				       | NONE => NONE)
			      | SOME [] => raise SympBug
				  ("ProofRules.MGU/reduce/checkCycle: found an empty cycle!")
			      | NONE => SOME C
			end handle NoUnifier => NONE)
		      | checkCycles [] = SOME []
		    (* Split tuple on the RHS when the LHS is of atomic type *)
		    fun splitTuple(C as ((m1, m2)::C')) =
			let fun main() =
				(case parts m2 of
				     [] => raise SympBug
					 ("ProofRules.MGU/reduce/splitTuple: can't happen")
				   | [ _ ] => [(C, S)]
				   | parts =>
				       (changed := true;
					List.foldr(op @)[]
					   (List.map(fn t=>reduce((m1, t)::C', S)) parts)))
			in
			    case messageType m1 of
				MessageType _ => [(C, S)]
			      | TupleType _ => raise SympBug
				    ("ProofRules.MGU/reduce/splitTuple: m1 is of type tuple: "
				     ^(pt2string m1))
			      | _ => main()
			end
		      | splitTuple [] = [([], S)]
		    (* Apply all the transformations in order. *)
		    val C = flatten C
		    val _ = debug(fn()=>funName^": flattened C = "^(C2str C)^"\n")
		    val C = tautology C
		    val _ = debug(fn()=>funName^": tautologies removed: C = "^(C2str C)^"\n")
		    val Copt = primType C
		    val _ = debug(fn()=>funName^": after primType: C = "
				  ^(case Copt of
					SOME C => (C2str C)
				      | NONE => "NONE")^"\n")
		    val Copt = Option.mapPartial checkCycles Copt
		    val _ = debug(fn()=>funName^": after checkCycles: C = "
				  ^(case Copt of
					SOME C => (C2str C)
				      | NONE => "NONE")^"\n")
		    (* This generates a list of new (C, S) pairs *)
		    val pairs = (case Copt of
				     NONE => []
				   | SOME C => splitTuple C)
		    val _ = debug(fn()=>funName^": after splitTuple: pairs = \n ["
				  ^(strlistmap2str ",\n  "
				    (fn(C, S)=>
				     "("^(subst2str S)^"\n  C = "^(C2str C)^")")
				    pairs)^"]\n")
		    (* Repeat the whole thing until C doesn't change,
		       or until C is unsatisfiable (then pairs are empty, 
		       but `changed' might not be set). *)
		    val res = 
			if !changed then List.foldr(op @) [] (List.map reduce pairs)
			else pairs
		    val _ = popFunStackLazy(funName, fn()=>
					    (strlistmap2str ",\n  "
					     (fn(C, S)=>
					      "("^(subst2str S)^"\n C = "^(C2str C)^")")
					     res))

		in res
		end
		
	    val res = reduce((m1, m2)::C, S)
	    val _ = popFunStackLazy(funName, fn()=>(strlistmap2str ",\n  "
						    (fn(C, S)=>
						     "("^(subst2str S)^"\n C = "^(C2str C)^")")
						    res))
	in res
	end

    (* MGUFilter returns only those bindings for which a most general
       unifier exists, and returns the appropriate substitution. *)

    fun MGUFilter C ([], _) = []
      | MGUFilter C (((binding as (Binding(_, h)))::t), goal as Goal(_, pt)) =
	  (List.map(fn(C,S)=>(S, C, binding)) (MGU (C, makeSubstitution()) (pt, h)))
	  @ (MGUFilter C (t, goal))

    (* Functions for carrying the substitution into the new state. *)
    fun subInNode S n =
	makeNode(nodeStrand n,
		 nodeIndex n,
		 nodeRole n,
		 Option.map (List.map (substitute S)) (nodeRoleParams n),
		 substitute S (nodeAction n))

    fun subInNodes (nodeHash, S) =
      let
	val newHash = makeHashDefault (nodeEq, nodeString)
	val allNodes = hash2any (fn x=>x) nodeHash
      in
	(List.app (fn (n, ()) =>
		   (insertHashDestructive (newHash, subInNode S n, ());
		    ()))
	 allNodes;
	 newHash)
      end

    fun subInRoles (roleHash, S) =
      let
	val newHash = makeHashDefault (op=, fn x=>x)
	val allRoles = hash2any (fn x=>x) roleHash
      in
	(List.app
	 (fn (s, role) =>
	  (insertHashDestructive (newHash, s, substitute S role);
	   ()))
	 allRoles;
	 newHash)
      end

    fun subInStrandSet S set = 
	  makeStrandSet(subInRoles(getRoles set, S),
			subInNodes(getNodes set, S))

    fun subInMessageExtr (ExtractMessage n, S) =
      ExtractMessage (subInNode S n)
      | subInMessageExtr (ExtractTuple (m, i), S) =
      ExtractTuple (subInMessageExtr (m, S), i)
      | subInMessageExtr (ExtractConcur (m, i), S) =
      ExtractConcur (subInMessageExtr (m, S), i)
      (* | subInMessageExtr (MTuple l, S) =
      MTuple (List.map (fn m => subInMessageExtr (m, S)) l) *)

    fun subInGoal (Goal(me, pt), S) =
      Goal(subInMessageExtr (me, S), substitute S pt)

    fun subInBinding (Binding(me, pt), S) =
      Binding(subInMessageExtr (me, S), substitute S pt)

    (* Substitution in the set of unbound goals.
       Take care of the goals that became non-atomic 
       (tuples) after the substitution. *)
    fun subInG (allGoals, S) =
      let
	(* Split tuples into separate goals.
	   Tuples may appear after the substitution. *)
	fun splitMsgExtr(m, MessageTuple(_, lst)) =
	    List.foldr(op @) [] (mapEnumerate(fn(i,pt)=>splitMsgExtr(ExtractTuple(m,i), pt)) lst)
	  | splitMsgExtr pair = [Goal pair]
	fun splitGoal(Goal(me, pt)) = splitMsgExtr(me, pt)
	fun addGoal g = splitGoal(subInGoal(g, S))
      in
	  List.foldr(op @) [] (List.map addGoal allGoals)
      end

    fun subInBindings (bindingHash, S) =
      let
	val newHash = makeHashDefault (goalEq, goalString)
	val allPairs = hash2any (fn x=>x) bindingHash
      in
	(List.app
	 (fn (g, b) =>
	  (insertHashDestructive (newHash, subInGoal (g, S),
				  subInBinding (b, S));
	   ()))
	 allPairs;
	 newHash)
      end

    (* Filter goals that ask for publicly known info, namely 
       pub. keys and names of principals. *)
    fun filterGoals goals =
	let fun relevant(Goal(_, Pub(_, k))) = not(isName k)
	      | relevant (Goal(_, x)) =
	        let val tp = messageType x
		in
		    case tp of
			PrincipalType _ => not(isName x)
		      | _ => true
		end
	in
	    List.filter relevant goals
	end

    (* sequentOf generates a new sequent and the instantiated current
       goal given a binding, goal, and substitution. *)

    fun sequentOf (sb, goalBinding, g, typeMap, delta, context, 
		   goal: Goal, binding: Binding, pruneParams, C, S) =
      let
	val state = {semibundle = subInStrandSet S sb,
		     goalBinding = subInBindings(insertHash (goalBinding, goal, binding), S),
		     intermConstr = C,
		     goals = filterGoals(subInG(List.filter(fn g'=> not(goalEq(g',goal))) g, S)),
		     typeMap = typeMap}: State
	val goal = subInGoal(goal, S)
      in
	(StateSequent {l = state, delta = subInStrandSet S delta, context = context,
		       pruneParams=pruneParams},
	 goal)
      end

    (* Add DESTRUCTIVELY a strand of the node `n' to the `rolesHash'.
       This is a hash mapping role Uids to hashs of strands.  Each
       strand hash maps a strand Uid to the strand's parameters. *)
    fun shovelNodeHash rolesHash n =
	(case findHash(rolesHash, nodeRole n) of
	     NONE => 
		 let val strands = makeHashDefault(op =, fn x=>x)
		     val _ = insertHashDestructive(strands, nodeStrand n,
						   nodeRoleParams n)
		 in (insertHashDestructive(rolesHash, nodeRole n, strands); ())
		 end
	   | SOME strands =>
		 (case findHash(strands, nodeStrand n) of
		      NONE => (insertHashDestructive(strands, nodeStrand n,
						     nodeRoleParams n); ())
		    | SOME _ => ()))

    (* Get the indices of `roleId' "new" parameters. *)
    fun getNewIndices (pHash: ProtocolHash) roleId =
	(case findHash(#newParams pHash, roleId) of
	     SOME lst => lst
	   | NONE => raise SympBug("ProofRules.getNewIndices: no such role: "^roleId))

    (* Get the indices of `roleId' "self" parameters. *)
    fun getSelfIndices (pHash: ProtocolHash) roleId =
	(case findHash(#selfParams pHash, roleId) of
	     SOME lst => lst
	   | NONE => raise SympBug("ProofRules.getSelfIndices: no such role: "^roleId))

    (* Given node `n', compute the list of predecessor nodes in the state. *)
    fun preimage(sb, goalBinding) n =
	let val (strand, index) = (nodeStrand n, nodeIndex n)
	    (* Find the nodes of the same strand with the preceding index *)
	    fun isPrevious(n,_)=
		if (nodeStrand n) = strand andalso
		    (nodeIndex n) = index-1 then SOME n
		else NONE
	    val prev = 
		if index = 1 then [] (* n is the first in the strand *)
		else List.mapPartial isPrevious (hash2any (fn x=>x) (getNodes sb))
	    (* Find all the nodes binding our goals *)
	    val localGoals = allGoals n (nodeAction n)
	    val bindings = List.mapPartial(fn g=>findHash(goalBinding, g)) localGoals
	    val binders = List.map Binding2Node bindings
	in prev@binders
	end
    (* Check:
     * For the cycles in the `<=-minimal' relation
     * That no two strands of the same role have the same fresh nonce (in state AND delta!)
     * That no two strands (from all the roles) have the same Self parameter
       (no magical `identity theft') *)

    fun isStateValid(protocolHash: ProtocolHash) (state: State, delta: StrandSet) =
	let val funName = "isStateValid"
	    exception BadState
	    val {semibundle = sb,
		 goalBinding=goalBinding,
		 goals=goals,
		 intermConstr=C, ...} = state
	    val {rolesUID=rolesUID,
		 penetratorUID=penetratorUID, ...} = protocolHash
	    val _ = pushFunStackLazy(funName, fn()=>"\n"^(state2stringDebug state))
	    val debug = lazyVerbDebug(getOptions()) funName
	    val getNewIndices = getNewIndices protocolHash
	    val getSelfIndices = getSelfIndices protocolHash
	    val preimage = preimage(sb, goalBinding)
	    (* Assign each node the set of terms that occur in it or before. *)
	    val termHash = makeHashDefault(nodeEq, nodeString)
	    (* map bindings to lists of goals (reverse the goalBinding) *)
	    val bindGoals = makeHashDefault(bindingEq, bindingString)
	    (* map nodes to their bindings *)
	    val bindHash = makeHashDefault(nodeEq, nodeString)
	    (* Initialize the mapping hashes *)
	    fun addBinding(g, b) =
		let val n = Binding2Node b
		    fun addHash(hash, k, v) =
			(insertHashDestructive(hash, k, 
					       (case findHash(hash, k) of
						    SOME lst => v::lst
						  | NONE => [v]));
			 ())
		in
		    (addHash(bindGoals, b, g);
		     addHash(bindHash, n, b))
		end		    
	    val _ = hash2any addBinding goalBinding

	    fun addTerms(n, terms) = (insertHashDestructive(termHash, n, terms); ())
	    (* Split message into atomic interms *)
	    fun parts(m as MessageTuple(_, lst)) = union ptEq (List.map parts lst)
	      | parts m = [m]

	    (* check if the node's action is to send a message *)
	    fun isSending(Send _) = true
	      | isSending(Labeled(_, p)) = isSending p
	      | isSending _ = false

	    (* Get the terms present in the node ***that bind some
               goals if node is sending*** and just all terms if
               receiving, then find all the interms from C and return
               them all together.  Forget the sent terms that no one
               receives. *)

	    fun getTerms n =
		let fun loop(Send(_, m)) = loop m
		      | loop(Receive(_, m)) = loop m
		      | loop(Labeled(_, p)) = loop p
		      | loop(MessageTuple(_, mlist)) = union ptEq (List.map loop mlist)
		      | loop(MessageConcur(_, mlist)) = union ptEq (List.map loop mlist)
		      | loop m = [m] (* Atomic message *)
		    (* For an atomic interm t, find all of its interms from C.
		       Assume that LHSs in C are already atomic. *)
		    fun interms t =
			let fun filter(t1,t2) =
			        if ptEq(t, t2) then SOME t1 else NONE
			in
			    union ptEq ([t]::(List.map interms (List.mapPartial filter C)))
			end
		    fun bindings n =
			(case findHash(bindHash, n) of
			     SOME l => l
			   | NONE => [])
		    fun bindingTerms(Binding(_, pt)) = loop pt
		    (* The immediate terms in the node *)
		    val terms =
			if isSending(nodeAction n) then
			    union ptEq (List.map bindingTerms (bindings n))
			else loop(nodeAction n)
		in
		    union ptEq (List.map interms terms)
		end

			  
	    (* Compute the terms for each node and check for <=-minimal relation *)
	    fun nodeTerms n =
		let val funName = funName^"/nodeTerms"
		    val _ = pushFunStackLazy(funName, fn()=>nodeStringVerbose n)
		    val res = 
			(case findHash(termHash, n) of
			     SOME terms => terms
			   | NONE =>
			     let val localTerms = getTerms n (* Terms in the node *)
				 (* Cycle detection: put what we already 
				    know about the term in the hash *)
				 val _ = insertHashDestructive(termHash, n, localTerms)
				 (* Terms in the preceding nodes (compute recursively) *)
				 val prevTerms = union ptEq (List.map nodeTerms (preimage n))
				 val res =
				     if isSending(nodeAction n) then
				      (case intersect ptEq [localTerms, prevTerms] of
					   [] => localTerms@prevTerms
					 | lst => (debug(fn()=>funName^": offending terms: ["
							 ^(strlistmap2str ", " pt2string lst)
							 ^"]\n in the node: "
							 ^(nodeStringVerbose n)^"\n");
						   raise BadState))
				     else union ptEq [localTerms, prevTerms]
				 val _ = insertHashDestructive(termHash, n, res)
			     in
				 res
			     end)
		    val _ = popFunStackLazy(funName, fn()=>strlistmap2str ", " pt2string res)
		in
		    res
		end

	    (* Check what it says, despite being so short and simple. *)	    
	    fun checkUniqueOrigination() =
		let val _ = hash2any(fn(n,_)=> nodeTerms n) (getNodes sb)
		in true
		end handle BadState => false

	    (* Checking fresh nonces.  The war plan:
	       - sort the nodes by roles;
	       - for each role, pick one node for each strand;

	       - for roles with >1 strand, find its FreshNonce
                 parameters, then see if any 2 strands have a common
                 instance of these parameters. *)

	    fun checkFreshParams() =
		let val funName = funName^"/checkFreshParams"
		    val _ = pushFunStack(funName, "starting...")
		    val debug = lazyVerbDebug(getOptions()) funName
		    (* map each regular roleId to the set of strands for that role *)
		    val rolesHash = makeHashDefault(op =, fn x=>x)
		    (* Same for penetrator roles *)
		    val pRoles = makeHashDefault(op =, fn x=>x)
		    (* Same for roles in Delta *)
		    val deltaHash = makeHashDefault(op =, fn x=>x)
		    (* Sef of all instances of `new' and `self' parameters.
		       Each parameter is mapped to the roleId (not strand)
		       that generates it. *)
		    val newHash = makeHashDefault(ptEq, pt2string)
		    (* Populate the role hashes *)
		    fun shovelNode (n,_) =
			(case findHash(penetratorUID, nodeRole n) of
			     SOME _ => shovelNodeHash pRoles n
			   | NONE => shovelNodeHash rolesHash n)
		    (* Only consider regular strands in Delta *)
		    fun shovelDeltaNode(n, _) =
			(case findHash(rolesUID, nodeRole n) of
			     SOME _ => shovelNodeHash deltaHash n
			   | NONE => ())
		    (* Now actually do it *)
		    val _ = hash2any shovelNode (getNodes sb)
		    val _ = hash2any shovelDeltaNode (getNodes delta)

		    fun strands2str strands =
			let fun strand2str(uid, SOME params) = 
			          uid^"["^(strlistmap2str ", " pt2string params)^"]"
			      | strand2str(uid, NONE) = uid
			in
			    "["^(strlist2str ",\n  " (hash2any strand2str strands))^"]"
			end

		    fun checkRole(roleId, strands) =
			let val funName = funName^"/checkRole"
			    (* Flag whether the role is regular or penetrator *)
			    val isRegular = not(isSome(findHash(penetratorUID, roleId)))
			    val _ = pushFunStackLazy(funName, fn()=>roleId^"("
						     ^(if isRegular then "regular"
						       else "peretrator")^"),\n "
						     ^(strands2str strands))
			    fun canonize(pt as UniqueId _) = pt
			      | canonize(TypedId(p, uid, id, _)) = UniqueId(p, uid, id)
			      | canonize pt = pt
			    (* List of parameter lists from all strands *)
			    val strandParams = hash2any(fn(_,SOME p)=> List.map canonize p
			                                | (_,NONE) => []) strands
			    (* Check that the i-th param is diff. from all the other 
			       elements in hash.  If `add' flag is true, add the new
			       parameters to the hash. *)
			    fun checkIndex (hash, add, self) i =
			        let val params = List.map (fn l=>List.nth(l, i)) strandParams
				         handle Subscript => raise SympBug
					   (funName^"/checkRole:\n "
					    ^" some strand for role "
					    ^roleId^" has too few parameters")
				    fun checkOne(p as UniqueId _) =
					(case findHash(hash, p) of
					     SOME rid =>
						 if self andalso rid = roleId then true
						 else
						  (debug(fn()=>funName
							 ^": Index check fails for role "
							 ^roleId^", parameter "^(Int.toString i));
						   false)
					   | NONE => ((if add then
							   insertHashDestructive(hash, p, roleId)
						       else hash);
						      true))
				      (* If it is not a variable, then it cannot be fresh *)
				      | checkOne _ = false
				in
				    List.all checkOne params
				end
			    (* Check that there are no exact clones of the same strand *)
			    fun checkForClones() =
				let val ptlist2str =
				          List.foldr(fn(pt, s)=>(pt2string pt)^","^s) ""
				    fun ptlistEq(l1, l2) =
					(case zipOpt(l1, l2) of
					     NONE => false
					   | SOME pairs => List.all ptEq pairs)
				    val hash = makeHashDefault(ptlistEq, ptlist2str)
				    fun checkOne params =
					(case findHash(hash, params) of
					     SOME _ => 
						 (debug(fn()=>funName
							^": Clone check fails for role ");
						  false)
					   | NONE => (insertHashDestructive(hash, params, ());
						      true))
				in 
				    List.all checkOne strandParams
				end
			    val res = (checkForClones() andalso
				        List.all (checkIndex(newHash, isRegular, false))
					         (getNewIndices roleId) andalso
				        List.all (checkIndex(newHash, isRegular, true))
					         (getSelfIndices roleId))
			    val _ = popFunStack(funName, if res then "true" else "false")
			in
			    res
			end (* checkRole *)

		    (* Check that the state does not violate identity and fresh parameter
		       conditions w.r.t. Delta.  That is, Self and other fresh parameters
		       are allowed to be the same in some state strand and some delta
		       strand only if their role is the same. *)

		    fun checkDelta() =
			let val funName = funName^"/checkDelta"
			    val _ = pushFunStack(funName, "starting")
			    val stateRoles = (hash2any(fn x=>x) pRoles)
			                    @(hash2any(fn x=>x) rolesHash)
			    val deltaRoles = hash2any(fn x=>x) deltaHash

			    fun checkStateStrand(sameRole, newParams, stateNewIndices)
				                (sid, params) =
			        let val stateNewParams =
				        (case params of
					     SOME p => 
						 List.map(fn i=>List.nth(p, i)) stateNewIndices
					   | NONE => [])
					handle Subscript => raise SympBug
					   (funName^"/checkStateStrand:\n "
					    ^" strand "^sid^" has too few parameters")
				    (* Check if two ParseTree lists are disjoint *)
				    fun disjoint(l1, l2) =
					(case intersect ptEq [l1, l2] of
					     [] => true
					   | _ => false)
				in
				    sameRole orelse
				      (disjoint(newParams, stateNewParams))
				end

			    fun checkStateRole(rid, []) _ = true (* nothing to check *)
			      | checkStateRole(rid, newParams) (stateRid, strandsHash) =
				let val strands = hash2any(fn x=>x) strandsHash
				    val stateNewIndices = (getNewIndices stateRid)
					                 @(getSelfIndices stateRid)
				    val checkStateStrand = 
					checkStateStrand(rid=stateRid, newParams, stateNewIndices)
				in
				    List.all checkStateStrand strands
				end
			    fun checkDeltaStrand (rid, newIndices) (sid, params) =
				let val newParams =
				        (case params of
					     SOME p => 
						 List.map(fn i=>List.nth(p, i)) newIndices
					   | NONE => [])
					handle Subscript => raise SympBug
					   (funName^"/checkDeltaStrand:\n "
					    ^" strand "^sid^" for role "
					    ^rid^" has too few parameters")
				in
				    List.all (checkStateRole(rid, newParams)) stateRoles
				end
			    fun checkDeltaRole(rid, strandsHash) =
				let val strands = hash2any(fn x=>x) strandsHash
				    val newIndices = (getNewIndices rid)
					            @(getSelfIndices rid)
				in
				    List.all (checkDeltaStrand(rid, newIndices)) strands
				end
			    val res = List.all checkDeltaRole deltaRoles
			    val _ = popFunStack(funName, if res then "true" else "false")
			in
			    res
			end
		    val res = (List.all (fn x=>x) (hash2any checkRole rolesHash))
			     andalso (List.all (fn x=>x) (hash2any checkRole pRoles))
			     andalso checkDelta()
		    val _ = popFunStack(funName, if res then "true" else "false")
		in
		    res
		end (* checkFreshParams *)

	    val res = checkUniqueOrigination() andalso checkFreshParams()
	    val _ = popFunStackLazy(funName, fn()=>(if res then "true" else "false"))
	in
	    res
	end (* isStateValid *)


    (* Constructs the next sequents for the split rule. *)

    fun nextSequents printFun ((roleHash, nodeHash), goalBinding, C, goals,
			       goalNumOpt, typeMap, delta, context,
			       protocolHash, pruneParams) =
      let val funName = "nextSequents"
	  val _ = pushFunStackLazy(funName, fn()=>"context = ["
				   ^(strlistmap2str ", " pt2string context)^"]")
	  val debug = lazyVerbDebug(getOptions()) funName
	  val getNewIndices = getNewIndices protocolHash
	  val getSelfIndices = getSelfIndices protocolHash
	  (* Add a role (both parse tree and nodes) to the current
	     state's semibundle (non-destructively). *)
	  fun addRole(roleOpt, newNodeHash) =
	      let val funName = funName^"/addRole"
		  val _ = pushFunStackLazy(funName, fn()=>
					   (case roleOpt of
						 SOME (name, r) => "\n role = "^name
					       | NONE => "NONE"))
		  val debug = lazyVerbDebug(getOptions()) funName
		  val nh = copyHash nodeHash
		  

		  val rh = (case roleOpt of
				SOME(name, roleDef) => insertHash(roleHash, name, roleDef)
			      | NONE => roleHash)
		  val _ = if sameHash(newNodeHash, nodeHash) then ()
			  else
			      (List.app(fn (n,_) => (insertHashDestructive(nh, n, ()); ()))
			               (hash2any(fn x=>x) newNodeHash))
		  val _ = popFunStackLazy(funName, fn()=>"\n "
					  ^(strlistmap2str "\n " nodeString (hash2any #1 nh)))
	      in
		  (rh, nh): SemiBundle
	      end
	  (* Add new unbound goals from the `newNodeHash' *)
	  fun addNewGoals(goals, newNodeHash) =
	      if sameHash(newNodeHash, nodeHash) then goals
	      else removeRedundant goalEq ((sortList goalLess (generateGoals newNodeHash))@goals)

	  (* Make a new node hash with only the nodes relevant to the binding,
	     and that not already present in the current `nodeHash'. *)
	  fun rebuildNodes(newNodeHash, binding) =
	      let val hash = wipeHash newNodeHash (* New empty hash of the same type *)
		  (* Reindex nodes by their numbers *)
		  val numHash = makeHashDefault(op =, Int.toString)
		  val _ = hash2any(fn(n, _)=>
				   (case findHash(nodeHash, n) of
					SOME _ => ()
				      | NONE => (insertHashDestructive(numHash, nodeIndex n, n);
						 ())))
		                  newNodeHash
		  val node = Binding2Node binding
		  val maxNodeNum = nodeIndex node
		  val relevantNodes =
		      if maxNodeNum <= 0 then []
		      else List.mapPartial(fn x=>x)
			  (List.tabulate(maxNodeNum, fn i=>findHash(numHash, i+1)))
		  val _ = List.app(fn n=>(insertHashDestructive(hash, n, ());
					  ())) relevantNodes
	      in hash
	      end
	  (* Try to bind the goal to the nodes in a given role.
	     Generate new sequents/hints. *)
	  fun doRole(goal, roleOpt, newNodeHash) =
	      let val funName = funName^"/doRole"
		  val debug = lazyVerbDebug(getOptions()) funName
		  val { rolesUID = rolesUID,
		        penetratorUID = penetratorUID, ...} = protocolHash
		  fun nodeHash2str nh =
		      let val lst = hash2any(fn(n,_) => nodeString n) nh
		      in
			  strlist2str ", " lst
		      end
		  fun roleOpt2str NONE = ""
		    | roleOpt2str(SOME(name, role)) = ",\n role ("^name^") = "^(pt2string role)
		  fun bind2str lst = ("["^(strlistmap2str ", " bindingString lst)^"]")
		  val _ = pushFunStackLazy(funName, fn()=>"goal = "^(goalString goal)
					   ^(roleOpt2str roleOpt)^",\n nodes = ["
					   ^(nodeHash2str newNodeHash)^"]")
		  val bindings = findBindings newNodeHash
		  val _ = debug(fn()=>funName^": bindings = "^(bind2str bindings)^"\n")
		  val subBindingList = MGUFilter C (bindings, goal)
		  val _ = debug(fn()=>funName^": filtered bindings = "
				^(bind2str(List.map #3 subBindingList))^"\n")

		  (* Match the strands in state with delta under the current
		     substitution, and generate new list of bindings/substitutions.

		     If a state strand shares some "new" parameters with a delta strand,
                     then all `new' *and* `self' parameters of those strands must be
		     unified. *)

		  fun matchDelta(S, C, binding) =
		      let val newNodeHash = subInNodes(rebuildNodes(newNodeHash, binding), S)
			  val newSemiBundle = subInStrandSet S (addRole(roleOpt, newNodeHash))
			  val delta = subInStrandSet S delta
			  (* map each regular roleId to the set of strands for that role *)
			  val rolesHash = makeHashDefault(op =, fn x=>x)
			  (* Same for roles in Delta *)
			  val deltaHash = makeHashDefault(op =, fn x=>x)
			  (* Populate the role hashes *)
			  fun shovelNode (n,_) = shovelNodeHash rolesHash n
			  (* Only consider regular strands in Delta *)
			  fun shovelDeltaNode(n, _) =
			      (case findHash(rolesUID, nodeRole n) of
				   SOME _ => shovelNodeHash deltaHash n
				 | NONE => ())
			  (* Now actually do it *)
			  val _ = hash2any shovelNode (getNodes newSemiBundle)
			  val _ = hash2any shovelDeltaNode (getNodes delta)
			  (* Check if two ParseTree lists are disjoint *)
			  fun disjoint(l1, l2) =
			      (case intersect ptEq [l1, l2] of
				   [] => true
				 | _ => false)
			  (* See if any role in delta has any common unique params with 
			     any strand of the same role in the state.  If it does,
			     unify all the other unique params (but only unique, not the others).*)
			  fun matchStrands _ [] = NONE
			    | matchStrands(args as (uniqueIndices, selfIndices))
			                  ([(sid1, params1), (sid2, params2)]::strands) =
			      let val (newParams1, selfParams1,
				       newParams2, selfParams2) =
				       (case (params1, params2) of
					    (NONE, NONE) => ([], [], [], [])
					  | (SOME p1, SOME p2) =>
						(List.map(fn i=>List.nth(p1, i)) uniqueIndices,
						 List.map(fn i=>List.nth(p1, i)) selfIndices,
						 List.map(fn i=>List.nth(p2, i)) uniqueIndices,
						 List.map(fn i=>List.nth(p2, i)) selfIndices)
					  | _ => raise SympBug
					     (funName^
					      "/matchStrands: params don't match for strands "
					      ^sid1^" and "^sid2))
				  val uniqueParams1 = newParams1@selfParams1
				  val uniqueParams2 = newParams2@selfParams2
				  val substs = 
				      if disjoint(newParams1, newParams2)
					  orelse List.all ptEq (zip(uniqueParams1, uniqueParams2))
				      then matchStrands args strands
				      else SOME(MGU(C,S) (MessageTuple(dp, uniqueParams1),
							  MessageTuple(dp, uniqueParams2)))
			      in
				  substs
			      end
			    | matchStrands _ _ = raise SympBug
			       (funName^"/matchStrands: This can't happen")
						 
			  fun matchDeltaRoles [] = NONE
			    | matchDeltaRoles ((rid, strandsHash)::roles) =
			      let val uniqueIndices = getNewIndices rid
				  val selfIndices = getSelfIndices rid
				  val h2a = hash2any(fn x=>x)
				  val strands = h2a strandsHash
			      in
				  case findHash(rolesHash, rid) of
				      NONE => matchDeltaRoles roles
				    | SOME stateStrandsHash =>
					  (case matchStrands(uniqueIndices, selfIndices)
					         (comblist[strands, h2a stateStrandsHash]) of
					       NONE => matchDeltaRoles roles
					     | substs => substs)
			      end
		      in
			  case matchDeltaRoles(hash2any(fn x=>x) deltaHash) of
			      NONE => [(S, C, binding)]
			    | SOME lst => List.foldr(op @) []
				  (List.map(fn(C,S)=>matchDelta(S,C,binding)) lst)
		      end (* matchDelta *)
		  fun makeSequent(S, C, binding) =
		      let val newNodeHash = rebuildNodes(newNodeHash, binding)
			  val newSemiBundle = addRole(roleOpt, newNodeHash)
		      in
			      sequentOf (newSemiBundle, goalBinding,
					 addNewGoals(goals, newNodeHash),
					 typeMap, delta, context, goal, binding,
					 pruneParams, C, S)
		      end
		  val subBindingList = List.foldr(op @) [] (List.map matchDelta subBindingList)
		  val res = List.map makeSequent subBindingList
		  val _ = popFunStackLazy(funName, fn()=>strlistmap2str "\n  "
					  (fn(seq,_)=>"["^(sequent2string seq)^"]") res)
	      in
		  res
	      end (* doRole *)
	  (* Try to bind the goal to new roles.  Generate sequents/hints. *)
	  fun matchRoles goal =
	      let val funName = funName^"/matchRoles"
		  val debug = lazyVerbDebug(getOptions()) funName
		  val { roles = roles,
		        rolesUID = rolesUID,
		        penetratorUID = penetratorUID,
		        penetrator=penetrator, ...} = protocolHash

		  fun getRole roleId =
		      let val hashes = [rolesUID, penetratorUID]
			  fun loop [] = raise SympBug
			        (funName^"/getExistingStrands: role "
				 ^roleId^" is undefined")
			    | loop (hash::hashes) =
				(case findHash(hash, roleId) of
				     SOME r => r
				   | NONE => loop hashes)
		      in
			  loop hashes
		      end

		  (* Given the list of triples (strandId, roleId, paramsOpt),
		     generate the node hash of all corresponding complete strands *)
		  fun makeNewNodeHash lst =
		      let val hash = makeHashDefault(nodeEq, nodeString)
			  fun addStrand(strandId, roleId, paramsOpt) =
			      let val (role, pa) = extractRoleParams(getRole roleId)
				  val role' = instantiate(role, pa, paramsOpt)
				  val body = extractBody role'
				  fun addNode(num, action) = 
				      let val node = makeNode(strandId, num+1, roleId, 
							      paramsOpt, action)
				      in
					  (insertHashDestructive(hash, node, ()); ())
				      end
			      in
				  (mapEnumerate addNode body; ())
			      end
		      in
			  (List.app addStrand lst; hash)
		      end
				  
		  (* Generate semibundle hashes for the role (given as
		     roleUid and the full parse tree for the role).
		     If the optional strand is given -- a pair
		     (strandId, paramsOpt), then convert the role into
		     that complete strand. *)

		  fun processRole strandOpt (name, role as Role(p, {name=n,
								    params=params,
								    body=body})) =
		      let fun newUid(TypedId(p,_,id,tp)) = TypedId(p, genSym id, id, tp)
			    | newUid x = raise SympBug
			       (funName^": parameter is not TypedId: "^(pt2stringVerbose x))
			  fun newVars lst =
			      let fun param2lst(Parameter(_, lst, _)) = lst
				    | param2lst x = [x]
			      in List.map newUid (List.foldr(op @) [] (List.map param2lst lst))
			      end
			  fun makeStrandTriple lst =
			      let val roleUid = getUid n
				  val (strandId, params) =
				    (case strandOpt of
					 NONE => (genSym roleUid, Option.map newVars lst)
				       | SOME pair => pair)
			      in
				  (strandId, roleUid, params)
			      end
			  val newNodeHash = makeNewNodeHash[makeStrandTriple params]
			  val _ = debug(fn()=>funName^": processing role: "
					^(pt2stringVerbose role)^"\n")
		      in doRole(goal, SOME(name, body), newNodeHash)
		      end
		    | processRole _ (name, x) = raise SympBug
		        (funName^": not a role: name = "^name
			 ^",\n expr = "^(pt2stringVerbose x))
		  (* Generate the list of strands (roles with params) in the current state.
		     Returns list of triples (strandId, roleId, paramsOpt) *)
		  fun getExistingStrands() =
		      let val hash = makeHashDefault(op =, fn x=>x)
			  val nodes = sortList nodeLess (hash2any #1 nodeHash)
			  fun doNode n =
			      (case findHash(hash, nodeStrand n) of
				   SOME _ => NONE
				 | NONE =>
				       (insertHashDestructive(hash, nodeStrand n, ());
					SOME(nodeStrand n, nodeRole n, nodeRoleParams n)))
		      in
			  List.mapPartial doNode nodes
		      end
		  fun strandSeq(strandId, roleId, params) =
		        processRole (SOME(strandId, params)) (roleId, getRole roleId)
		  val existingRoleSeq = List.map strandSeq (getExistingStrands())
		  val roleSeq = List.map (processRole NONE) (hash2any(fn x=>x) rolesUID)
		  val penetrSeq = List.map (processRole NONE) (hash2any(fn x=>x) penetratorUID)
	      in
		  List.foldr(op @) [] (existingRoleSeq @ roleSeq @ penetrSeq)
	      end (* matchRoles *)

	  (* Filter out those sequents to which the `Decr' pruning
	     theorem applies.  Basically, a state is invalid if it has
	     a goal with encryption depth higher than Emax.

	     `origGoal' is the goal before instantiation, and each
	     sequent is paired with the same goal instantiated for its
	     binding. *)

	  fun pruneFilter(SOME Emax) (seqList, origGoal) =
	      let val funName = "pruneFilter"
		  val _ = pushFunStack(funName, "starting")
		  fun goalEmax C (Goal(_, pt)) = maxEncryptionDepth (SOME C) pt
		  fun loop(seq as StateSequent{l = state, ...}, goal) =
		      let val funName = funName^"/loop"
			  val _ = pushFunStackLazy(funName, fn()=>sequent2string seq)
			  val goals = #goals state
			  val C = #intermConstr state
			  val Elist = List.map (goalEmax C) goals
			  val _ = lazyVerbDebug(getOptions()) funName
			        (fn()=>funName^": Emax for goals = ["
				 ^(strlistmap2str "," Int.toString Elist)^"]\n")
			  val Edepth = maxList(op <, 0) Elist
			  val res = Edepth <= Emax
			  val _ = popFunStack(funName, if res then "true" else "false")
		      in
			  res
		      end
		    | loop(InitSequent _, _) = false
		  val res = List.filter loop seqList
		  val _ = popFunStack(funName, "[\n ["
				      ^(strlistmap2str "]\n  [" (sequent2string o #1) res)
				      ^"]]")
	      in
		  res
	      end
	    | pruneFilter NONE (seqList, _) = seqList		      

	      (* Compute the `importance' of a goal.
	      
	         Currently, the importance is the number of *bound*
	         goals that are subterms of the current one.  The
	         rationale for that is, the more goals the current
	         goal will satisfy, the better. *)

(*  	  fun bindingDepth *)

	  (* Pick a goal to bind.  It should be more intelligent in
	     the future.  For instance, we may try to bind all the
	     goals, and pick the one with no bindings (which would
	     prove the current sequent).  Or we may analyze the
	     respective nodes and pick the lowest ones. *)

	  fun pickGoal(goals as g::_) =
	      (case goalNumOpt of
		   SOME n => (List.nth(goals, n-1) handle Subscript => raise ProverError
		       ("No such goal in the sequent"))
		 | NONE => g)
	    | pickGoal [] = raise NoGoals (* This case shouldn't happen anymore *)

	  (* Check that the state in the sequent is valid (not contradictory) *)
	  fun isValid(StateSequent{l = state, delta=delta, ...}, _) =
	        isStateValid protocolHash (state, delta)
	    | isValid(InitSequent_, _) = false

	  val goal = pickGoal goals
	  val _ = printFun("\nBinding the goal "^(goalString goal)^"\n")

	  (* Find the goal that has valid bindings and generate new sequents for it. *)
	  val newSeqs = (matchRoles goal)       (* Add new roles *)
	  fun res2str res () =
	        strlistmap2str "\n  " (fn(seq,_)=>"["^(sequent2string seq)^"]") res
	  val _ = debug(fn()=>funName^": sequents before filtering: ["
			^(res2str newSeqs ())^"]\n")
	  val res = List.filter isValid newSeqs
	  val res = pruneFilter (#Emax pruneParams) (res, goal)
	  val res = List.map(fn(s,_)=>(s, NONE: SequentHints option)) res
	  val _ = popFunStackLazy(funName, res2str res)
      in
	  res
      end

    (* Calls the appropriate functions for the Athena "split"
       rule. *)

    fun splitApply printFun (StateSequent 
			     {l = {semibundle = sb,
				   goalBinding = goalBinding,
				   intermConstr = C,
				   goals = goals,
				   typeMap = map}, 
			      delta = delta, context = context,
			      pruneParams=pruneParams},
			     pcontext, args, substOpt) =
	let val funName = "splitApply"
	    val debug = lazyVerbDebug(getOptions()) funName
	    val _ = debug(fn()=>funName^": context = "^(ProgramContext2str pcontext)^"\n")
	    val protocolHash = (case pcontext of
				    Local(_, (_, phash)) => phash
				  | Global _ => raise SympBug
				     (funName^": wrong context (Global instead of Local):\n"
				      ^(ProgramContext2str pcontext)))
	    val goalNum = (case findIRArgFormulaNumValue args "goal" of
			       SOME (GoalPart, n) => n
			     | SOME _ => raise ProverError
				   ("Wrong sequent part in the goal number.")
			     | NONE => NONE)
	in 
	    if List.length goals = 0 then InferenceRuleFailure
	    else 
		((InferenceRuleSuccess 
		  (nextSequents printFun (sb, goalBinding, C, goals, goalNum, map,
					  delta, context, protocolHash, pruneParams)))
		 handle NoGoals => InferenceRuleFailure)
	end
      | splitApply printFun _ = InferenceRuleFailure

    fun applyPruneKeys printFun
	  (StateSequent{l = {semibundle = sb,
			     goalBinding = goalBinding,
			     intermConstr = C,
			     goals = goals,
			     typeMap = map}, 
			delta = delta, context = context,
			pruneParams=pruneParams},
	   pcontext, args, substOpt) =
	  let val (sym, priv) = (#symKeySent pruneParams, #privKeySent pruneParams)
	      fun isKey(Goal(_, pt)) = isSubtype(messageType pt, KeyType dp)
	      fun isSymKey(Goal(_, pt)) = isSubtype(messageType pt, SymKeyType dp)
	      fun isPrivKey(Goal(_, pt)) = isSubtype(messageType pt, PrivKeyType dp)
	  in
	      case (sym, priv) of
		  (true, true) => InferenceRuleFailure
		| (true, false) => if List.exists isPrivKey goals then InferenceRuleSuccess []
				   else InferenceRuleFailure
		| (false, true) => if List.exists isSymKey goals then InferenceRuleSuccess []
				   else InferenceRuleFailure
		| (false, false) => if List.exists isKey goals then InferenceRuleSuccess []
				    else InferenceRuleFailure
	 end
      | applyPruneKeys _ _ = InferenceRuleFailure
	    
    fun applyPruneSigned printFun (seq, pcontext, args, substOpt) =
	 let val funName = "applyPruneSigned"
	     val _ = pushFunStackLazy(funName, fn()=>sequent2string seq)
	     val debug = lazyVerbDebug(getOptions()) funName
	     fun isSigned goal =
		 let val res = (case goal of
				    (Goal(_, Crypt(_, _, Priv _))) => true
				  | _ => false)
		     val _ = debug(fn()=>funName^": "^(goalString goal)
				   ^(if res then " is a signed message\n"
				     else " is not a signed message\n"))
		 in res
		 end
	     val res = (case seq of
			    StateSequent{l = {goals = goals, ...},
					 pruneParams=pruneParams, ...} =>
			       if #signedSent pruneParams then InferenceRuleFailure
			       else if List.exists isSigned goals then InferenceRuleSuccess []
			       else InferenceRuleFailure
			   | _ => InferenceRuleFailure)
	     val _ = popFunStack(funName, (case res of
					       InferenceRuleFailure => "Fails"
					     | _ => "Succeeds"))
	 in
	     res
	 end

    (* The Athena "init" rule transforms a formula and protocol into
       a state sequent.  The formula is first changed into CNF, then
       rearranged as a series of conjunctions of implications.  These
       become the state sequent subtrees.  The strands in the antecedent
       and succedent of the implications are placed into gamma and delta,
       respectively.  In the initial states, no goals are bound.  A
       search is then performed of nodes to find goals that are unbound. 
       The closure within roles is performed at this stage.  *)
 
    val init = {name = "init",
		help = "Initializes the Athena sequent.",
		params = [],
		match = SOME (fn _ => SOME()),
		apply = SOME initApply}

    (* The final rule applies to leaf sequents.  A leaf sequent is a
       sequent where delta has a non-empty intersection with the state
       l.  We define intersection as meaning some strand in delta is
       a substrand of some strand in the state l. *)

    val final = {name = "final",
		 help = "Applies to leaf sequents.",
		 params = [],
		 match = SOME (fn _ => SOME()),
		 apply = SOME finalApply}

    (* The split rule selects an unbound goal and creates subtrees for
       each possible binding for it.   It does not apply if there are
       no unbound goals. *)

    val split = {name = "split",
		 help = "Attempts to split an Athena state sequent.",
		 params = [{name="goal",
			    Type=ProverArgFormulaNumType}],
		 match = SOME (fn _ => SOME()),
		 apply = SOME splitApply}

    val pruneKeysRule =
	{name = "prune_keys",
	 help = ("If no regular strand in the protocol sends out a key in any form, and\n"
		 ^"there is an unbound goal in the state of a subtype of Key, then this\n"
		 ^"goal can never be satisfied, and the state is contradictory (with the\n"
		 ^"default intruder).  Since we have a special DecrSign intruder role\n"
		 ^"which extracts signed messages, binding a key variable to someone's\n"
		 ^"public key is not necessary for soundness.\n\n"
		 ^"This rule will check all the above conditions and prove the current\n"
		 ^"subgoal if it satisfies them."),
	 params = [],
	 match = SOME (fn _ => SOME()),
	 apply = SOME applyPruneKeys}: InferenceRule

    val pruneSignedRule =
	{name = "prune_signed",
	 help = ("If no regular strand in the protocol  signs messages with its private key, and\n"
		 ^"there is an unbound goal in the state receiving a signed message, then this\n"
		 ^"goal can never be satisfied, and the state is contradictory (with the\n"
		 ^"default intruder).\n\n"
		 ^"This rule will check all the above conditions and prove the current\n"
		 ^"subgoal if it satisfies them."),
	 params = [],
	 match = SOME (fn _ => SOME()),
	 apply = SOME applyPruneSigned}: InferenceRule

    (* Command to print the current sequent for debugging *)
    fun printDebugApply printFun (seq, args) =
	  (printFun((sequent2stringDebug seq)^"\n"); SOME ())

    val printDebugCommand = {name = "print_debug",
			     help = "Print the current sequent in an extremely verbose way.",
			     params = [],
			     apply = printDebugApply}: ProofSystemCommand

    (* Command to print the current sequent for debugging *)
    fun bindingsApply printFun (StateSequent{l = { goalBinding = binding,
						   intermConstr = C, ...}, ...}, args) =
	(printFun("\n"^(binding2string(binding, C))^"\n"); SOME ())
      | bindingsApply printFun (InitSequent _, args) = NONE

    val bindingsCommand = {name = "bindings",
			     help = "Print the current goal bindings in the current sequent.",
			     params = [],
			     apply = bindingsApply}: ProofSystemCommand

    (* Given a directed acyclic graph as a list of nodes, a hash of
       edges (mapping each node to the list of the next nodes), and
       equality/string conversion functions for nodes, sort the nodes
       topologically and return them in a list in the ascending order,
       where the "smaller" nodes are the earlier ones in the graph.

       Rases SympBug if the graph is cyclic. *)

    fun orderGraphNodes(nodes, edges, nodeEq, node2str) =
	let val edges = copyHash edges (* make a clean copy *)
	    (* Same edges in the reverse direction *)
	    val backEdges = wipeHash edges
	    fun addBackEdge(n1, lst) =
		let fun doOne n2 =
		        (case findHash(backEdges, n2) of
			     SOME lst => (insertHashDestructive(backEdges, n2, n1::lst); ())
			   | NONE => (insertHashDestructive(backEdges, n2, [n1]); ()))
		in
		    List.app doOne lst
		end
	    val _ = hash2any addBackEdge edges
	    fun getNext(edges, n) = (case findHash(edges, n) of
				      SOME lst => lst
				    | NONE => [])
	    val visited = makeHashDefault(nodeEq, node2str)
	    val ordered = ref []

	    (* Do DFS on the graph, and insert the nodes into
	       `ordered' list in the post order: first, left subtree,
	       then right, then the node itself, and then backtrack. *)
	    fun dfs n =
		(case findHash(visited, n) of
		     SOME _ => ()
		   | NONE =>
		      let val next = getNext(edges, n)
		      in
			  (insertHashDestructive(visited, n, ());
			   List.app dfs next;
			   ordered := n::(!ordered))
		      end)
	    val initNodes = List.filter(fn n=>not(isSome(findHash(backEdges, n)))) nodes
	in
	    (List.app dfs initNodes;
	     !ordered)
	end

    fun explainApply printFun (StateSequent{l = state, delta=delta, ...}, args) =
	let val { semibundle = sb,
		  goalBinding = goalBinding,
		  intermConstr = C, ...} = state
	    val preimage = preimage(sb, goalBinding)
	    val nodeHash = getNodes sb
	    (* map bindings to lists of goals (reverse the goalBinding) *)
	    val bindGoals = makeHashDefault(bindingEq, bindingString)
	    (* map nodes to their bindings *)
	    val bindHash = makeHashDefault(nodeEq, nodeString)
	    (* Set of strands, mapping a strand ID to some node on that strand *)
	    val strandHash = makeHashDefault(op =, fn x=>x)
	    val deltaStrandHash = makeHashDefault(op =, fn x=>x)
	    val edges = makeHashDefault(nodeEq, nodeString)
	    (* Initialize the mapping hashes *)
	    fun addBinding(g, b) =
		let val n = Binding2Node b
		    fun addHash(hash, k, v) =
			(insertHashDestructive(hash, k, 
					       (case findHash(hash, k) of
						    SOME lst => v::lst
						  | NONE => [v]));
			 ())
		in
		    ((case findHash(bindGoals, b) of
			  SOME _ => () (* This binding is already added *)
			| NONE => addHash(bindHash, n, b));
		     addHash(bindGoals, b, g))
		end
	    fun addStrand hash n =
		(case findHash(hash, nodeStrand n) of
		     SOME _ => ()
		   | NONE => (insertHashDestructive(hash, nodeStrand n, n); ()))
	    (* Representing state as a graph for topological sorting *)
	    fun addEdge n = (case preimage n of
				 [] => ()
			       | lst => (insertHashDestructive(edges, n, lst); ()))

	    val _ = hash2any addBinding goalBinding
	    val nodes = hash2any #1 nodeHash
	    val _ = List.app addEdge nodes
	    val _ = List.app (addStrand strandHash) nodes
	    val _ = List.app (addStrand deltaStrandHash) (hash2any #1 (getNodes delta))
	    val ordered = List.rev(orderGraphNodes(nodes, edges, nodeEq, nodeString))
	    (* If node is sending, compose a list of tuples (node, recipients, message),
	       and skip receiving nodes (return empty list). *)
	    fun node2actions n =
		let fun binding2action(b as Binding(_, message)) =
		        (case findHash(bindGoals, b) of
			     SOME [] => NONE
			   | NONE => NONE
			   | SOME goals => SOME(n, List.map Goal2Node goals, message))
		in
		    case findHash(bindHash, n) of
			NONE => []
		      | SOME bindings => List.mapPartial binding2action bindings
		end
	    fun action2str(node, recip, message) =
		(nodeString node)^" -> "
		^(strlistmap2str ", " nodeString recip)
		^": "^(pt2string message)^"\n"
	    fun strand2str node =
		(nodeStrand node)^(case nodeRoleParams node of
				       SOME lst => "["^(strlistmap2str ", " pt2string lst)^"]"
				     | NONE => "")^"\n"
	    fun C2str C =
		"["^(strlistmap2str ",\n "
		     (fn(t,t')=>"("^(pt2string t)^" << "^(pt2string t')^")") C)
		^"]"
	    val actions = List.foldr(op @) [] (List.map node2actions ordered)
	    val _ = printFun "\n\n Protocol run:\n\n"
	    val _ = List.app (printFun o action2str) actions
	    val _ = (case C of
			 [] => ()
		       | _ => printFun ("\n Interm Constraints:\n\n"^(C2str C)^"\n"))
	    val _ = printFun "\n Participants:\n\n"
	    val _ = List.app (printFun o strand2str) (sortList nodeLess (hash2any #2 strandHash))
	    val _ = printFun "\n Strands in Delta:\n\n"
	    val _ = List.app (printFun o strand2str) (hash2any #2 deltaStrandHash)
	    val _ = printFun "\n"
	in
	    SOME ()
	end
      | explainApply printFun (InitSequent _, args) =
	(printFun "\nThis is an initial sequent, which should be self-explanatory.\n";
	 NONE)

    val explainCommand = {name = "explain",
			  help = "Print an intelligible protocol run for the current sequent.",
			  params = [],
			  apply = explainApply}: ProofSystemCommand

    fun applyTactic _ _ _ _ = NONE

    fun proofRules () = [init, final, split, pruneKeysRule, pruneSignedRule]

    fun proverCommands () = [printDebugCommand, bindingsCommand, explainCommand]

  end
