functor AthenaProofSystemFun(structure Interface: INTERFACE
			     structure Hash: HASH): PROOF_SYSTEM =
  struct

    structure ProofRules = 
      ProofRulesAthenaFun(structure Hash = Hash
			  structure Interface = Interface)

    open ProofRules
    open Interface
    open Sequent
    open SequentCommon
    open SequentBase
    open Hash
    open SympBug
    open Str

    val debugNames = 
	[(* Typechecker *)
	 "checkMessage",
	 "checkTheorem",
	 "checkSend",
	 "checkReceive",
	 "knowKey",
	 "getNewIndices",
	 "Athena:typeCheckProgram",
	 "addPenetrator",
	 "Athena:typeCheckProtocol",
	 (* Sequent *)
	 "maxEncryptionDepth",
	 (* Proof rules *)
	 "ProofRules.prepare",
	 "ProofRules.CNF",
	 "ProofRules.dist",
	 "subStrand",
	 "ProofRules.substitute",
	 "getProtocolPruneParams",
	 "MGU",
	 "MGU/loop",
	 "MGU/reduce",
	 "MGU/reducesInvKey",
	 "isStateValid",
	 "isStateValid/nodeTerms",
	 "isStateValid/checkFreshParams",
	 "isStateValid/checkFreshParams/checkRole",
	 "isStateValid/checkFreshParams/checkDelta",
	 "pruneFilter",
	 "pruneFilter/loop",
	 "splitApply",
	 "applyPruneSigned",
	 "nextSequents",
	 "nextSequents/addRole",
	 "nextSequents/doRole",
	 "nextSequents/matchRoles"]

    local

      val rules = makeHashDefault(op =, fn x=>x): (string, InferenceRule) Hash

      val commands = makeHashDefault(op =, fn x=>x) : 
	(string, ProofSystemCommand) Hash

      val tactics = makeHashDefault(op =, fn x=>x): 
	(string, ProofSystemTactic) Hash

    in

      fun findRuleByName str = findHash(rules, str)
      fun findCommandByName str = findHash(commands, str)
      fun findTacticByName str = findHash(tactics, str)
      fun addNewRule r = (insertHashDestructive(rules, getRuleName r, r); ())
      fun addNewCommand c = 
	(insertHashDestructive(commands, getCommandName c, c); ())
      fun addNewTactic t = 
	(insertHashDestructive(tactics, getTacticName t, t); ())
	
      (* Initialize the set of inference rules from the ProofRules module *)
      fun initProofSystem () = 
	(List.app addNewRule (proofRules());
	 List.app addNewCommand (proverCommands()))
      fun getAllRules() = hash2any(fn (_,x) => x) rules
      fun getAllCommands() = hash2any(fn (_,x) => x) commands
      fun getAllTactics() = hash2any(fn (_,x) => x) tactics
	
      fun getInitRule() = 
	(case findRuleByName "init" of
	   SOME r => r
	 | NONE => raise SympBug
	     ("ProofSystemDefault/getInitRule: no \"init\" rule"))

      (* Apply the rule, and check if the resulting sequents are trivially
       true by applying the init rule.  All sequents that are proven this way
       are discarded from the result. *)

      fun applyRule printFun (seq, context) (rule : InferenceRule, 
					     args, substOpt) = 
	(case (#apply rule) of
	   SOME apply => 
	     (case substOpt of
		SOME subst =>
		  apply printFun (seq, context, args, subst)
	      | NONE =>
		  apply printFun (seq, context, args, ()))
	 | NONE =>
		InferenceRuleFailure)
	   
      fun applyCommand printFun (seq, context) (command: ProofSystemCommand, args) =
	  ((#apply command) printFun (seq, args)): ProofSystemCommandResult option
	
      (* Must return SOME(ruleName, args, newSequentsList, newTacticState) or 
       NONE *)
      fun applyTactic inline printFun (tactic, args, context, stateOpt)
	(seq, hintsOpt) =

(*
    type ProofSystemTactic = unit
    type ProofSystemTacticState = unit

    fun initProofSystem() = ()

    fun findRuleByName name = NONE: InferenceRule option
    fun findCommandByName name = NONE: ProofSystemCommand option
    fun findTacticByName name = NONE: ProofSystemTactic option

    fun matchRule (seq, context) (rule, args) = (NONE: Substitution option)
    fun applyRule printFun (seq, context) (rule, args, substOpt) = InferenceRuleFailure
    fun applyCommand printFun (seq, context) (command, args) =
	  NONE: ProofSystemCommandResult option
*)
    (* Must return SOME(ruleName, args, newSequentsList, newTacticState) or NONE *)
(*    
fun applyTactic inline printFun (tactic, args, context, stateOpt)
                    (seq, hintsOpt) =
*)
	NONE: (string * (InferenceRuleArgument list)
	       * ((Sequent * (SequentHints option)) list)
	       * (ProofSystemTacticState option)) option
	
    end

  end
