signature ATHENA_TYPES =
  sig

    structure Hash: HASH
    structure ParseTreeStruct: ATHENA_PARSE_TREE

    (* A ProtocolHash contains mappings from strings (names) to 
       the roles, penetrators, predicates and theorems in the protocol. *)

    type ProtocolHash  = {(* Roles indexed by name *)
			  roles : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash,
			  (* Same roles indexd by UniqueID *) 
			  rolesUID : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash,
			  (* The ID of the protocol used as the penetrator *)
			  penetratorProtocol: string,
			  (* Penetrator roles by name *)
			  penetrator : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash,
			  (* Penetrator roles by UniqueID *)
			  penetratorUID : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash,
			  (* `New' parameters of all roles, indexed by roleUID *)
			  newParams: (string, int list) Hash.Hash,
			  (* `Self' parameters of all roles, indexed by roleUID *)
			  selfParams: (string, int list) Hash.Hash,
			  predicates : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash,
			  theorems : (string, ParseTreeStruct.ParseTree) 
			  Hash.Hash} 

    (* Context for typechecking *)
    datatype ProgramContext = 
      Global of (string, ParseTreeStruct.ParseTree * ProtocolHash) Hash.Hash |
      Local of string * (ParseTreeStruct.ParseTree * ProtocolHash)

    (* Typechecked program, expression, declaration, whatever, which
       is the result of typechecking a ParseTree. *)   
    datatype ProgramObject = 
      ProgramHash of (string, ProtocolHash) Hash.Hash |
      Object of ParseTreeStruct.ParseTree

    val ProtocolHash2str: ProtocolHash -> string
    val ProgramContext2str: ProgramContext -> string
    val ProgramObject2str: ProgramObject -> string

    (* Generate a new unique ID.  This function belongs to some common
       place, so that IDs are truely unique to all the modules. *)

    (* Generate a completely new name *)
    val nextUID: unit -> string
    (* Generate a unique name based on the given string *)
    val genSym: string -> string

  end

