functor AthenaTypesFun(structure Hash: HASH
		       structure ParseTreeStruct: ATHENA_PARSE_TREE):
  ATHENA_TYPES =
  struct

    structure Hash = Hash
    structure ParseTreeStruct = ParseTreeStruct

    open Hash
    open ParseTreeStruct
    open Str

    type ParseTree = ParseTreeStruct.ParseTree

    type ProtocolHash = {(* Roles indexed by name *)
			 roles : (string, ParseTree) Hash,
			 (* Same roles indexd by UniqueID *) 
			 rolesUID: (string, ParseTree) Hash,
			 (* The ID of the protocol used as the penetrator *)
			 penetratorProtocol: string,
			 (* Penetrator roles by name *)
			 penetrator : (string, ParseTree) Hash,
			 (* `New' parameters of all roles, indexed by roleUID *)
			 newParams: (string, int list) Hash,
			 (* `Self' parameters of all roles, indexed by roleUID *)
			 selfParams: (string, int list) Hash,
			 (* Penetrator roles by UniqueID *)
			 penetratorUID : (string, ParseTree) Hash,
			 predicates : (string, ParseTree) Hash,
			 theorems : (string, ParseTree) Hash}
      
    datatype ProgramContext = 
      Global of (string, ParseTree * ProtocolHash) Hash |
      Local of string * (ParseTree * ProtocolHash)
      
    datatype ProgramObject = 
      ProgramHash of (string, ProtocolHash) Hash |
      Object of ParseTree

    (* Functions for debugging printing *)
    fun ProtocolHash2str(phash: ProtocolHash) =
	let val {roles=roles,
		 penetrator=penetrator,
		 newParams=newParams,
		 selfParams=selfParams,
		 predicates=predicates,
		 theorems=theorems, ...} = phash
	    fun hash2str h =
		let fun doOne(k, v) = k^" -> "^(pt2string v)
		in
		    "["^(strlist2str ", " (hash2any doOne h))^"]"
		end
	in
	    "ProtocolHash{\n  roles = "^(hash2str roles)
	    ^",\n  penetrator = "^(hash2str penetrator)
	    ^",\n  newParams = ["
	    ^(strlist2str ", " (hash2any(fn(r, l)=> "("^r^" -> ["
					    ^(strlistmap2str "," Int.toString l)^"])")
				   newParams))^"]"
	    ^",\n  selfParams = ["
	    ^(strlist2str ", " (hash2any(fn(r, l)=> "("^r^" -> ["
					    ^(strlistmap2str "," Int.toString l)^"])")
				   selfParams))^"]"
	    ^",\n  predicates = "^(hash2str predicates)
	    ^",\n  theorems = "^(hash2str theorems)^"}"
	end

    fun ProgramContext2str(Global hash) =
	let fun pair2str(name, (pt, phash)) =
	        "Protocol "^name^" = [\n  "^(pt2string pt)
		^"\n"^(ProtocolHash2str phash)^"]\n"
	in
	    "Global(\n"^(strlist2str "\n\n" (hash2any pair2str hash))
	    ^"\n)\n"
	end
      | ProgramContext2str(Local(name, (pt, phash))) =
	 "Local(Protocol "^name^" = [\n  "^(pt2string pt)
	 ^"\n"^(ProtocolHash2str phash)^"])\n"

    fun ProgramObject2str(ProgramHash hash) =
	let fun pair2str(name, phash) = 
	         "Protocol "^name^" = [\n  "^(ProtocolHash2str phash)^"]\n"
	in
	    "ProgramHash(\n"^(strlist2str "\n\n" (hash2any pair2str hash))^"\n)\n"
	end
      | ProgramObject2str(Object pt) = "Object("^(pt2string pt)^")"


    (* Generate a new unique ID.  This function belongs to some common
       place, so that IDs are truely unique to all the modules. *)

    local
      val c = ref 0 
    in
      fun nextUID () =
	(("!" ^ (Int.toString (!c))) before (c := (!c) + 1))

      fun genSym id = id ^ (nextUID())
    end



  end
