(* Built-in default penetrator for Athena *)
functor DefaultPenetratorFun(structure ParseTreeStruct: ATHENA_PARSE_TREE): PENETRATOR =
  struct
    structure ParseTreeStruct = ParseTreeStruct

    open ParseTreeStruct
    open Pos

    val penetratorName = "DefaultPenetrator"

    val penetrator =
	Protocol((0,1), Id((0,2), penetratorName),
		 [Role((~1,3), {name=Id((~1,4), "Flush"),
			    params=SOME[Parameter((~1,5), [Id((~1,6), "g")],
						  MessageType (~1,7))],
			    body=RoleBody((~1,8), [Receive((~1,9), Id((~1,10), "g"))])}),
		  (* These roles will never be used *)
		  (* Role(dp, {name=Id(dp,"Tee"),
			    params=SOME[Parameter(dp, [Id(dp, "g")], MessageType dp)],
			    body=RoleBody(dp, [Receive(dp, Id(dp, "g")),
					       Send(dp, MessageConcur(dp, [Id(dp, "g"),
									   Id(dp, "g")]))])}),
		  (* Role V *)
		  Role(dp, {name=Id(dp,"Concat"),
			    params=SOME[Parameter(dp, [Id(dp, "g"), Id(dp, "h")],
						  MessageType dp)],
			    body=RoleBody(dp, [Receive(dp, MessageConcur(dp, [Id(dp, "g"),
									      Id(dp, "h")])),
					       Send(dp, MessageTuple(dp, [Id(dp, "g"),
									  Id(dp, "h")]))])}),

		  (* Role R *)
		  Role(dp, {name=Id(dp,"Split"),
			    params=SOME[Parameter(dp, [Id(dp, "g"), Id(dp, "h")],
						  MessageType dp)],
			    body=RoleBody(dp, [Receive(dp, MessageTuple(dp, [Id(dp, "g"),
									     Id(dp, "h")])),
					       Send(dp, MessageConcur(dp, [Id(dp, "g"),
									   Id(dp, "h")]))])}),
                  *)
		  Role((~5,3), {name=Id((~5,4),"Encr"),
			    params=SOME[Parameter((~5,5), [Id((~5,6), "k")], KeyType (~5,7)),
					Parameter((~5,8), [Id((~5,9), "h")],
						  MessageType (~5,10))],
			    body=RoleBody((~5, 11), [Receive((~5, 12), 
							     MessageConcur((~5, 13), 
									   [Id((~5, 14), "k"),
									    Id((~5, 15), "h")])),
						     Send((~5, 16), Crypt((~5, 17),
									  Id((~5, 18), "h"),
									  Id((~5, 19), "k")))])}),
		  (* Sign messages with penetrator's key *)
		  Role((~6, 1), {name=Id((~6,2),"Encr2"),
			    params=SOME[Parameter((~6, 3), [Id((~6, 4), "E")], 
						  New((~6,5), PrincipalType (~6, 5))),
					Parameter((~6, 6), [Id((~6, 7), "h")],
						  MessageType (~6, 8))],
			    body=RoleBody((~6, 9), [Receive((~6, 10), Id((~6, 11), "h")),
					       Send((~6, 12), Crypt((~6, 13), Id((~6, 14), "h"),
							      Priv((~6, 15), Id((~6, 16), "E"))))])}),
		  Role((~7, 1), {name=Id((~7,2),"DecrSym"),
			    params=SOME[Parameter((~7, 3), [Id((~7, 4), "k")], SymKeyType (~7, 5)),
					Parameter((~7, 6), [Id((~7, 7), "h")],
						  MessageType (~7, 8))],
			    body=
			     let val encrM = Crypt((~7, 9), Id((~7, 10), "h"),
						   Id((~7, 12), "k"))
			     in RoleBody((~7, 13), 
					 [Receive((~7, 14), Id((~7, 16), "k")),
					  Receive((~7, 15), encrM),
					  Send((~7, 17), Id((~7, 18), "h"))])
			     end}),
		  Role((~7, 1), {name=Id((~7,2),"Decr"),
			    params=SOME[Parameter((~7, 3), [Id((~7, 4), "A")], PrincipalType (~7, 5)),
					Parameter((~7, 6), [Id((~7, 7), "h")],
						  MessageType (~7, 8))],
			    body=
			     let val encrM = Crypt((~7, 9), Id((~7, 10), "h"),
						   Pub((~7, 11), Id((~7, 12), "A")))
			     in RoleBody((~7, 13), 
					 [Receive((~7, 14), Priv((~7, 42), Id((~7, 16), "A"))),
					  Receive((~7, 15), encrM),
					  Send((~7, 17), Id((~7, 18), "h"))])
			     end}),
		  (* Receive messages encrypted with penetrator's key *)
		  Role((~8, 1), {name=Id((~8,2),"Decr2"),
			    params=SOME[Parameter((~8, 3), [Id((~8, 4), "D")], 
						  New((~8,5), PrincipalType (~8, 5))),
					Parameter((~8, 6), [Id((~8, 7), "h")],
						  MessageType (~8, 8))],
			    body=
			     let val encrM = Crypt((~8, 9), Id((~8, 10), "h"), Pub((~8, 11), Id((~8, 12), "D")))
			     in RoleBody((~8, 13), [Receive((~8, 14), encrM),
					      Send((~8, 15), Id((~8, 16), "h"))])
			     end}),
		  (* Receive messages signed with someone's private key and dercrypt them *)
		  Role((~9, 1), {name=Id((~9,2),"DecrSign"),
			    params=SOME[Parameter((~9, 3), [Id((~9, 4), "A")], PrincipalType (~9, 5)),
					Parameter((~9, 6), [Id((~9, 7), "h")],
						  MessageType (~9, 8))],
			    body=
			     let val encrM = Crypt((~9, 9), Id((~9, 10), "h"), Priv((~9, 11), Id((~9, 12), "A")))
			     in RoleBody((~9, 13), [Receive((~9, 14), encrM),
					      Send((~9, 15), Id((~9, 16), "h"))])
			     end}),
		  (* Make penetrator do something useful: 
		     Supply the public key of any participant *)
		  Role((~10, 1), {name=Id((~10, 2), "PubKeys"),
			    params=SOME[Parameter((~10, 3), [Id((~10, 4), "A")], PrincipalType (~10, 5))],
			    body=RoleBody((~10, 6), [Send((~10, 7), Pub((~10, 8), Id((~10, 9), "A")))])}),

		  (* Send all symmetric keys the penetrator might share with others.
		     Currently, we only allow exactly two participants to share the same key. *)
		  Role((~10, 1), {name=Id((~10, 2), "SymKeys"),
			    params=SOME[Parameter((~10, 3), [Id((~10, 4), "D")],
						  New((~10, 42), PrincipalType (~10, 5))),
					Parameter((~10, 3), [Id((~10, 4), "A")],
						  PrincipalType (~10, 5))],
			    body=
			     let val key1 = SymKey((~10, 44), [Id((~10, 4), "D"),
							       Id((~10, 4), "A")])
				 val key2 = SymKey((~10, 45), [Id((~10, 4), "A"),
							       Id((~10, 4), "D")])
				  
			     in RoleBody((~10, 6), 
					 [Send((~10, 7),
					       MessageConcur((~10, 43), [key1, key2]))])
			     end}),

		  (* The penetrator can also send random noise to the network: his
		   own fresh nonces *)
		  Role((~11, 1), {name=Id((~11,2), "Nonces"),
			    params=SOME[Parameter((~11, 3), [Id((~11, 4), "ND")],
						  New((~1,5), NonceType (~11, 5)))],
			    body=RoleBody((~11, 6), [Send((~11, 7), Id((~11, 8), "ND"))])})
		  ]) (* end of Protocol *)

  end
