signature INPUT_LANGUAGE_DEFAULT =
  sig

    exception ParseError of string
    exception TypeError of string

    structure Typecheck: TYPECHECK
    structure Options: OPTIONS
    sharing Options = Typecheck.Evaluate.ParserDefault.Interface.Options
    type options = Options.options

(*    structure Evaluate: EVALUATE = Typecheck.Evaluate
    structure ParserDefault: PARSER_DEFAULT = Typecheck.Evaluate.ParserDefault
*)
    type ObjectHash = Typecheck.ObjectHash
    type ParseTree = Typecheck.Evaluate.ParserDefault.ParseTreeStruct.ParseTree

    (* Typechecked program, expression, declaration, whatever, which
       is the result of typechecking a ParseTree. *)
    type ProgramObject (* = {obj: ParseTree, objectHash: ObjectHash} *)
    (* Type or an expression*)
    type ExpressionType (* = {Type: ParseTree, objectHash: ObjectHash} *)
    (* Context for typechecking *)
    type ProgramContext (* = {context: ParseTree, objectHash: ObjectHash} *)

    datatype SelectObject =
	SelectObject of ProgramObject
      | SelectType of ExpressionType
      | SelectContext of ProgramContext

    (* String values for keywords starting a context ("module",
     "protocol", etc.), a theorem (usually "theorem", an expression
     (e.g. "val"), and a type (e.g. "type") *)

    val contextKeyword: string
    val theoremKeyword: string 
    val expressionKeyword: string
    val typeKeyword: string

    val Parse: options -> TextIO.instream -> ParseTree
    val ParseInput: options -> TextIO.instream -> ParseTree
    val ParseFile: options -> string -> ParseTree
    val wasError: bool ref

    (* Typechecking *)

    (* Typecheck the entire program, return the top-level context *)
    val typeCheckProgram: options -> ParseTree -> ProgramContext

    (* Typecheck a construct that represents some context (e.g. a
       module).  Take the top context and the parse tree of the
       construct and return the new (local) context. *)
    val typeCheckContext: options -> ProgramContext -> ParseTree
	  -> ProgramContext

    (* Typecheckng an expression and return its new typechecked
       representation together with its type *)
    val typeCheckExpr: options -> ProgramContext -> ParseTree * ParseTree
	      -> ProgramObject * ExpressionType

    (* Same for a theorem *)
    val typeCheckTheorem: options -> ProgramContext -> ParseTree
	  -> ProgramObject

    (* Add a new object, context, or type to the object hash *DESTRUCTIVELY*.
       Returns the uid of the new object.
       The uid can be optionally specified as the second parameter. *)
    val addObject: ProgramObject * (ParseTree option) -> ParseTree
    val addContext: ProgramContext * (ParseTree option) -> ParseTree
    val addType: ExpressionType * (ParseTree option) -> ParseTree

    (* Find and classify an object in the object hash.  The key must be a Uid. *)
    val findSelectObject: ObjectHash * ParseTree -> SelectObject option

    (* Fonce some of the objects to be ProgramContext.  (The one above
       treats some contexts like funClosure as objects) *)
    val findProgramContext: ObjectHash * ParseTree -> ProgramContext option
    (* Just to be complete and fair *)
    val findProgramObject: ObjectHash * ParseTree -> ProgramObject option
    val findExpressionType: ObjectHash * ParseTree -> ExpressionType option

    (* Equality on parse trees and objects *)
    val ptEq: ParseTree * ParseTree -> bool
    val poEq: ProgramObject * ProgramObject -> bool
    val pcEq: ProgramContext * ProgramContext -> bool

  (* Additional functions *)

    val ProgramObject2pt: ProgramObject -> ParseTree
    val ProgramObject2ObjectHash: ProgramObject -> ObjectHash
    val ProgramContext2pt: ProgramContext -> ParseTree
    val ProgramContext2ObjectHash: ProgramContext -> ObjectHash
    val makeProgramObject: ParseTree * ObjectHash -> ProgramObject
    val makeProgramContext: ParseTree * ObjectHash -> ProgramContext
    val makeExpressionType: ParseTree * ObjectHash -> ExpressionType

  end
