signature PROOF_RULES_DEFAULT_COMMON =
  sig
    structure Sequent: SEQUENT_DEFAULT

    type ParseTree = Sequent.SequentCommon.SequentBase.InputLanguage.ParseTree
    type AsstVarsTree = Sequent.SequentCommon.SequentBase.Trans.AsstVarsTree
    type Model = Sequent.Model
    type TermID = Sequent.TermID
    type TermHash = Sequent.TermHash
    type Sequent = Sequent.Sequent
    type Abstraction = Sequent.SequentCommon.SequentBase.Trans.Abstract.Abstraction
    type TypeAbstraction = Sequent.SequentCommon.SequentBase.Trans.Abstract.TypeAbstraction

    type ReplaceFlags = {prop: bool, initOnly: bool, RHSonly: bool}

    val isNotProp: ParseTree -> bool
    (* Check whether the expression has temporal operators in it *)
    val hasTemporal: ParseTree -> bool
    (* Check if all temporal operators in the expression are variants
       of X (next time op.), and X's are not nested. *)
    val nextOnly: ParseTree -> bool

    val isACTL: ParseTree -> bool
    val isECTL: ParseTree -> bool
    val isLTL: ParseTree -> bool
    (* Branching time mu-calculus with only universal path quantifiers.
       CTL operators are also allowed. *)
    val isAMu: ParseTree -> bool
    (* Branching time mu-calculus with only existential path quantifiers.
       CTL operators are also allowed. *)
    val isEMu: ParseTree -> bool
    (* Linear time mu-calculus.  LTL operators are also allowed. *)
    val isLTMu: ParseTree -> bool

    (* If `prop' is true, do not substitute inside temporal operators
       replaceFormula prop (term, old, new) *)
    val replaceFormula: bool -> ParseTree * ParseTree * ParseTree -> ParseTree

    (* replaceAVT (initOnly, RHSonly) (tree, old, new) *)
    val replaceAVT: bool * bool -> AsstVarsTree * ParseTree * ParseTree -> AsstVarsTree
    (* replaceModel (initOnly, RHSonly) (model, old, new) *)
    val replaceModel: bool * bool -> Model * ParseTree * ParseTree -> Model

    (* replaceSeq flags (seq, old, new) *)
    val replaceSeq: ReplaceFlags -> Sequent * ParseTree * ParseTree -> Sequent
    val updateSeqModel: Model * Sequent -> Sequent
    (* Replaces the abstraction in the model by the new abstraction *)
    val updateSeqAbstraction: Abstraction * Sequent -> Sequent
    (* Adds a new type abstraction to the model of the sequent *)
    val addSeqTypeAbstraction: TypeAbstraction * Sequent -> Sequent

    (* Statically evaluate the term given by its `id', convert back to
       the termID representation and return it.
        evalTermID findObject termHash id*)
    val evalTermID: (ParseTree -> ParseTree option) -> TermHash -> TermID -> TermID
  end
