functor ProofRulesDefaultCommon(structure Sequent: SEQUENT_DEFAULT): PROOF_RULES_DEFAULT_COMMON =
  struct

    structure Sequent = Sequent

    open Sequent
    open SequentCommon
    open SequentBase
    open InputLanguage
    open Typecheck.Evaluate
    open Trans
    open ParserDefault.ParseTreeStruct
    open ParserDefault.Interface
    open Options

    type ReplaceFlags = {prop: bool, initOnly: bool, RHSonly: bool}

    fun isNotProp (Builtin{name=name,...}) = isNotProp name
      | isNotProp (Object{def=def,...}) = isNotProp def
      | isNotProp (MuClosure _) = true
      | isNotProp (NuClosure _) = true
      | isNotProp x = isTemporal x

    (* Check whether the expression has temporal operators in it *)
    fun hasTemporal pt =
	if isTemporal pt then true
	else List.exists hasTemporal (ptSubtrees pt)

    (* Check if all temporal operators in the expression are variants
       of X (next time op.), and X's are not nested. *)
    fun nextOnly(Ax _) = true
      | nextOnly(Ex _) = true
      | nextOnly(NextTime _) = true
      | nextOnly(Builtin{name=name,...}) = nextOnly name
      | nextOnly(Object{def=def,...}) = nextOnly def
      | nextOnly(Appl(_, ff, arg)) =
	  if nextOnly ff then not(hasTemporal arg) else true
      | nextOnly pt = if isTemporal pt then false else List.all nextOnly (ptSubtrees pt)	

    fun isACTL(Ax _) = true
      | isACTL(Ag _) = true
      | isACTL(Af _) = true
      | isACTL(Au _) = true
      | isACTL(Ar _) = true
      | isACTL(Builtin{name=name,...}) = isACTL name
      | isACTL(Object{def=def,...}) = isACTL def
      | isACTL pt = if isTemporal pt then false else List.all isACTL (ptSubtrees pt)

    fun isECTL(Ex _) = true
      | isECTL(Eg _) = true
      | isECTL(Ef _) = true
      | isECTL(Eu _) = true
      | isECTL(Er _) = true
      | isECTL(Builtin{name=name,...}) = isECTL name
      | isECTL(Object{def=def,...}) = isECTL def
      | isECTL pt = if isTemporal pt then false else List.all isECTL (ptSubtrees pt)

    fun isLTL(NextTime _) = true
      | isLTL(Globally _) = true
      | isLTL(Eventually _) = true
      | isLTL(Until _) = true
      | isLTL(Releases _) = true
      | isLTL(Builtin{name=name,...}) = isLTL name
      | isLTL(Object{def=def,...}) = isLTL def
      | isLTL pt = if isTemporal pt then false else List.all isLTL (ptSubtrees pt)

    (* Branching time mu-calculus with only universal path quantifiers.
       CTL operators are also allowed. *)
    fun isAMu(Ax _) = true
      | isAMu(Ag _) = true
      | isAMu(Af _) = true
      | isAMu(Au _) = true
      | isAMu(Ar _) = true
      | isAMu(Mu _) = true
      | isAMu(Nu _) = true
      | isAMu(Builtin{name=name,...}) = isAMu name
      | isAMu(Object{def=def,...}) = isAMu def
      | isAMu pt = if isTemporal pt then false else List.all isAMu (ptSubtrees pt)

    (* Branching time mu-calculus with only existential path quantifiers.
       CTL operators are also allowed. *)
    fun isEMu(Ex _) = true
      | isEMu(Eg _) = true
      | isEMu(Ef _) = true
      | isEMu(Eu _) = true
      | isEMu(Er _) = true
      | isEMu(Mu _) = true
      | isEMu(Nu _) = true
      | isEMu(Builtin{name=name,...}) = isEMu name
      | isEMu(Object{def=def,...}) = isEMu def
      | isEMu pt = if isTemporal pt then false else List.all isEMu (ptSubtrees pt)

    (* Linear time mu-calculus.  LTL operators are also allowed. *)
    fun isLTMu(NextTime _) = true
      | isLTMu(Globally _) = true
      | isLTMu(Eventually _) = true
      | isLTMu(Until _) = true
      | isLTMu(Releases _) = true
      | isLTMu(Mu _) = true
      | isLTMu(Nu _) = true
      | isLTMu(Builtin{name=name,...}) = isLTMu name
      | isLTMu(Object{def=def,...}) = isLTMu def
      | isLTMu pt = if isTemporal pt then false else List.all isLTMu (ptSubtrees pt)

    (* For a quick hack: transform everything to the ParseTree,
       replace as we usually do, and convert back to the termHash
       representation. *)

    (* If `prop' is true, do not substitute inside temporal operators *)
    fun replaceFormula prop (term, old, new) =
	let val funName = "replaceFormula"
	    val _ = pushFunStackLazy(funName, fn()=>"term="^(pt2stringDebug term)
				     ^", old="^(pt2stringDebug old)
				     ^", new="^(pt2stringDebug new))
	    fun recur t = ptTransform subst t
	    and subst t =
		let val funName = funName^"/subst"
		    val _ = pushFunStackLazy(funName, fn()=>(pt2stringDebug t))
		    val res = if ptEq(old, t) then new
			      else (case t of
					FunClosure{formals=f,...} => 
					    if eqUNames(f, old) then t else recur t
				      | Appl(_, Op, _) => 
						if prop andalso isNotProp Op then t else recur t
				      | _ => recur t)
		    val _ = popFunStackLazy(funName, fn()=>(pt2stringDebug res))
		in res
		end
	    val res = subst term
	    val _ = popFunStackLazy(funName, fn()=>(pt2stringDebug res))
	in res
	end
    fun replaceAVT (initOnly, RHSonly) (tree, old, new) =
	let fun subst e = replaceFormula false (e, old, new)
	    fun substVar(NormalVar v) = NormalVar(subst v)
	      | substVar(NextVar v) = NextVar(subst v)
	      | substVar(InitVar v) = InitVar(subst v)
	    fun substVars(vars as {init=initV, next=nextV, norm=normV}) =
		if RHSonly then vars else
		    let val initV = List.map substVar initV
			val nextV = if initOnly then nextV
				    else List.map substVar nextV
			val normV = if initOnly then normV
				    else List.map substVar normV
		    in {init=initV, next=nextV, norm=normV}
		    end			  
	    fun loop(t as NormalAsstTree(vars, v, e)) =
		if initOnly then t
		else NormalAsstTree(substVars vars,
				    if RHSonly then v else subst v,
				    subst e)
	      | loop(t as NextAsstTree(vars, v, e)) =
		if initOnly then t
		else NextAsstTree(substVars vars,
				  if RHSonly then v else subst v,
				  subst e)
	      | loop(t as InitAsstTree(vars, v, e)) = 
		InitAsstTree(substVars vars,
			     if RHSonly then v else subst v,
			     subst e)
	      | loop(NopAsstTree vars) = NopAsstTree(substVars vars)
	      | loop(ListAsstTree(vars, lst)) =
		ListAsstTree(substVars vars, List.map loop lst)
	      | loop(LetAsstTree(vars, decls, t)) =
		LetAsstTree(substVars vars, decls, loop t)
	      | loop(CaseAsstTree(vars, sel, pairs)) =
		let fun ff(u, p, t) = (u, subst p, loop t)
		in CaseAsstTree(substVars vars, subst sel, List.map ff pairs)
		end
	      | loop(IfAsstTree(vars, pairs, t)) =
		let fun ff(c, t) = (subst c, loop t)
		in IfAsstTree(substVars vars, List.map ff pairs, loop t)
		end
	      | loop(ChooseAsstTree(vars, paramsOpt, pairs)) =
		let fun ff(c, t) = (subst c, loop t)
		in ChooseAsstTree(substVars vars, paramsOpt, List.map ff pairs)
		end
	      | loop(ForeachAsstTree(vars, params, t)) =
		ForeachAsstTree(substVars vars, params, loop t)
	      | loop(LabeledAsstTree(vars, label, t)) =
		LabeledAsstTree(substVars vars, label, loop t)
	in loop tree
	end
    fun replaceModel (opt as (initOnly, RHSonly)) (model: Model, old, new) =
	let val {trans=trans,
		 objectHash=objectHash,
		 pvars=pvars,
		 abs=abs,...} = model
	    val globalPvars = ref(makePrimitiveVars())
	    val stateVars = ref []
	    val findObject = Typecheck.makeFindObject objectHash
	    fun replaceAtomicModel(AtomicModel{name=name,
					       uname=uname,
					       assts=assts,
					       cone=cone,
					       pvars=pvars,
					       absModules=absModules}) =
		let val options = getOptions()
		    val newAssts = replaceAVT opt (assts, old, new)
		    (* Rebuilding the tree may introduce vars that
		     were previously removed with the cone, so do
		     not use the actual tree, only cone and
		     pvars. *)
		    val opt = if initOnly then
			rebuildAsstOptionsDefault
			      else rebuildAsstOptionsBalance rebuildAsstOptionsDefault
		    val opt = rebuildAsstOptionsPreserveVars opt
		    val (_, newCone, newPvars) =
			rebuildAsstVars(options, findObject, pvars, (get_limit options), abs)
			(NONE, newAssts, opt)
		in
		    (* TODO: handle `absModules'; now they are left untouched *)
		    AtomicModel{name=name,
				uname=uname,
				assts=newAssts,
				cone=newCone,
				pvars=newPvars,
				absModules=absModules}
		end
	    fun replaceTrans (TransSync2(t1,t2)) =
		TransSync2(replaceTrans t1, replaceTrans t2)
	      | replaceTrans (TransAsync2(t1,t2)) =
		TransAsync2(replaceTrans t1, replaceTrans t2)
	      | replaceTrans (TransSync{names=names,
					body=t,
					parent=p}) =
		TransSync{names=names,
			  body=replaceTrans t,
			  parent=p}
	      | replaceTrans (TransAsync{names=names,
					 body=t,
					 parent=p}) =
		TransAsync{names=names,
			   body=replaceTrans t,
			   parent=p}
	      | replaceTrans (TransAtomic m) =
		let val newM as AtomicModel{pvars=newPvars,
					    assts=assts, ...} = replaceAtomicModel m
		    val {init=initvars, norm=normvars, next=nextvars} = getAsstVars assts
		    val vars = List.map vtName (normvars@initvars)
		in (globalPvars := mergePrimitiveVars(newPvars, !globalPvars);
		    stateVars := vars@(!stateVars);
		    TransAtomic newM)
		end
	in
	    {trans=replaceTrans trans,
	     objectHash=objectHash,
	     stateVars = !stateVars,
	     pvars = (!globalPvars),
	     abs=abs}: Model
	end

    fun replaceSeq (flags: ReplaceFlags) (seq: Sequent, old, new) =
	let val {prop=prop, initOnly=initOnly, RHSonly=RHSonly, ...} = flags
	    val { model=model,
		  init=initOpt,
		  invar=invar,
		  context=context,
		  assumptions=assump,
		  conclusions=conc } = seq
	    val newSeq = { model=replaceModel(initOnly, RHSonly)(model, old, new),
			   init=Option.map(fn x=>replaceFormula prop (x, old, new)) initOpt,
			   invar=List.map(fn x=>replaceFormula prop (x, old, new)) invar,
			   context=context,
			   assumptions=List.map(fn x=>replaceFormula prop (x, old, new)) assump,
			   conclusions=List.map(fn x=>replaceFormula prop (x, old, new)) conc }
	in newSeq
	end

    fun updateSeqModel(model: Model, seq: Sequent) =
	let val { init=initOpt,
		  invar=invar,
		  context=context,
		  assumptions=assump,
		  conclusions=conc, ... } = seq
	in { model=model,
	     init=initOpt,
	     invar=invar,
	     context=context,
	     assumptions=assump,
	     conclusions=conc }: Sequent
	end

    (* Replaces the abstraction in the model by the new abstraction *)
    fun updateSeqAbstraction(abs, seq) =
	let val { model=model, ... } = seq
	    val opt = rebuildAsstOptionsPreserveVars rebuildAsstOptionsDefault
	in updateSeqModel(rebuildModel opt (updateAbstraction(model, abs)), seq)
	end

    (* Adds a new type abstraction to the model of the sequent *)
    fun addSeqTypeAbstraction(typeAbs, seq) =
	let val { model=model, ... } = seq
	    val opt = rebuildAsstOptionsPreserveVars rebuildAsstOptionsDefault
	in 
	    updateSeqModel(rebuildModel opt (addTypeAbstraction(model, typeAbs)), seq)
	end



    (* Statically evaluate the term given by its `id', convert back to
       the termID representation and return it *)

    fun evalTermID findObject termHash id =
	let val pt = stripObjects(termID2pt termHash id)
	    val newPt = evaluateExpr(getOptions()) findObject pt
	in
	    pt2termID termHash newPt
	end

  end
