(* This is an abstract interface to the structure with basic type
   definitions for a proof system.  It is separate from the main
   Sequent module because we want to define other standard typed and
   functions for inference rules that use these types. *)

signature SEQUENT_BASE_Default =
  sig
    structure InputLanguage: INPUT_LANGUAGE_DEFAULT
    structure Trans: TRANS
    sharing InputLanguage.Typecheck = Trans.Typecheck

    type ParseTree = InputLanguage.ParseTree
    type ProgramContext = InputLanguage.ProgramContext

    (* Sequent: (M,I); Sigma; Gamma |- Delta *)
    type Sequent = { model: Trans.Model,
		     (* may or may not have the initial state predicate *)
		     init: ParseTree option,
		     invar: ParseTree list,
		     (* Context specific to the sequent.  In particular, 
		        all Skolem constants go here *)
		     context: ProgramContext,
		     assumptions: ParseTree list,
		     conclusions: ParseTree list }

    datatype SequentPart = InvarPart | AssumptionPart | ConclusionPart
      | AnyPart | SomeParts of SequentPart list

    (* Some hints of what a sequent should look like, so that subgoals
       in a proof tree can be matched better when the theorem changes.
       Currently, the hints are added only by the internal functions
       like rule application, etc. *)

    (* Some hints of what a sequent should look like, so that subgoals
       in a proof tree can be matched better when the theorem changes.
       Currently, the hints are added only by the internal functions
       like rule application, etc. *)

    (* Dummy type until we figure it out *)
    type SequentHints = unit

    (* Pretty printing *)
    val formulaInd2str: SequentPart * (int option) -> string
    val sequent2string: Sequent -> string
    val sequent2str: Sequent -> Str.Str

    val hints2string: SequentHints -> string

  end
