functor TransSMVFun (structure Hash : HASH
		     structure Trans: TRANS): TRANS_SMV =

  struct

    structure Hash = Hash
    structure Trans = Trans

    open ListOps
    open Hash
    open Trans
    open Str
    open Abstract
    open Pos
    open ParserDefault
    open ParseTreeStruct
    open Interface
    open Options

    type VarSMVMapping = (ParseTree, Str) Hash
    type DefineMapping = (ParseTree, Str) Hash
    type TypeValueMapping = (ParseTree, Str list) Hash

    (* Redefine exprDef to ignore the abstraction *)
    val exprDef = let val abs = makeAbstraction()
		  in fn findObject => expressionDef(findObject, abs, NONE)
		  end

    fun ptListEq ([], []) = true
      | ptListEq (x::l1, y::l2) =
	  if ptEq(x,y) then ptListEq(l1,l2)
	  else false
      | ptListEq _ = false

    (* Redefine name generator _not_ to include `!' symbols *)
    fun newName () = "X_"^(Int.toString(newNumber()))

    fun makeMapping () = makeHashDefault (ptEq, pt2string) 
    fun makeTVMapping () = makeHashDefault (ptEq, pt2string) 
    fun addMapDestructive(map, v, smv) = 
	(lazyVerbDebug (getOptions()) "TransSMV.addMap"
	  (fn()=>"\naddMap("^(pt2string v)^" -> "^(Str2string smv)^")\n");
	 insertHashDestructive(map, v, smv))

    fun addTVMapDestructive(map, v, smv) = 
	(lazyVerbDebug (getOptions()) "TransSMV.addTVMap"
	  (fn()=>"\naddTVMap("^(pt2string v)
	   ^" -> {"^(strlistmap2str ", "  Str2string smv)^"})\n");
	  insertHashDestructive(map, v, smv))

    fun map2str map =
	  "["^(strlist2str ",\n  "
	       (hash2any(fn(v,v')=>(pt2stringDebug v)^" -> "^(Str2string v')) map))
	  ^"]"

    fun TVMap2str map =
	"["^(strlist2str ",\n  "
	     (hash2any(fn(v,v')=>(pt2stringDebug v)^" -> "^(Str2string(Strlist2Str "," v'))) map))
	^"]"

    fun getMapOpt(map, v) =
	let val _ = pushFunStackLazy("TransSMV.getMapOpt",
				     fn()=>(pt2string v)^", "^(map2str map))
	    fun loop(v as Builtin{name=v', ...}) =
		(case findHash(map, v) of
		     SOME s => SOME s
		   | NONE => loop v')
	      | loop v = findHash(map, v)
	    val res = loop v
	    val _ = popFunStackLazy("TransSMV.getMapOpt",
				    fn()=>option2string "NONE" (Option.map Str2string res))
	in res
	end

    fun getMap(map, v) =
	(case getMapOpt(map, v) of
	     SOME s => s
	   | NONE => raise SympBug
		 ("trans_smv:  variable not in hash: "^(pt2stringDebug v)
		  ^"\n  varMap = "^(map2str map)))

    fun getTVMap(map, v) =
	let val _ = pushFunStackLazy("TransSMV.getTVMap",
				     fn()=>(pt2string v)^", "^(TVMap2str map))
	    val res = (case findHash (map, v) of
			   SOME s => s
			 | NONE => raise SympBug
			       ("trans_smv:  variable not in hash: "^(pt2stringDebug v)
				^"\n  TVMap = "^(TVMap2str map)))
	    val _ = popFunStackLazy("TransSMV.getTVMap",
				    fn()=>"["^(Str2string(Strlist2Str "," res))^"]")
	in res: Str list
	end

    fun isComplex (Appl (_, Builtin{name=And _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Or _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Implies _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Eq _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=NotEq _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Not _, ...}, _)) = true
      | isComplex x = false


    val bottom = Conc (Str (newName ()), Str "bot")
    val truth = Conc (Str (newName ()), Str "true")
    val falsity = Conc (Str (newName ()), Str "false")
    val eqTrue = Conc (Str " = ", truth)
    val eqFalse = Conc (Str " = ", falsity)
    val eqBottom = Conc (Str " = ", bottom)
    val thenTrue = Conc (Str " : ", Conc (truth, Str ";"))
    val thenFalse = Conc (Str " : ", Conc (falsity, Str ";"))
    val thenBottom = Conc (Str " : ", Conc (bottom, Str ";"))
    val otherwise = (Str "\n  1")
    val || = (Str " | ")
    val & = (Str " & ")
    val dummyBool = Conc (Str (newName ()), Str "ensureBoolean")

    (* Functions to translate 1- and 2-arity functions.
       There should be NO trailing `;' at the end. *)

    local fun op * (x,y) = Conc(x,y)
    in 
      fun codeFor1 (Not _, s) = (Str "!(") * s * (Str ")")
	| codeFor1 (Ag _, s) = (Str "AG(") * s * (Str ")")
	| codeFor1 (Af _, s) = (Str "AF(") * s * (Str ")")
	| codeFor1 (Eg _, s) = (Str "EG(") * s * (Str ")")
	| codeFor1 (Ef _, s) = (Str "EF(") * s * (Str ")")
	| codeFor1 (Ax _, s) = (Str "AX(") * s * (Str ")")
	| codeFor1 (Ex _, s) = (Str "EX(") * s * (Str ")")
	| codeFor1 (Object {def = d, ...}, s) = codeFor1 (d, s)
	| codeFor1 (ObjectInst {obj = d, ...}, s) = codeFor1 (d, s)
	| codeFor1 (Builtin {name=n, ...}, s) = codeFor1 (n, s)
	| codeFor1 (x, s) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-1 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s)^")")
	  
      fun codeFor2 (And _, s1, s2) =
	   (Str "(") * s1 * (Str ")")* & * (Str "(") * s2 * (Str ")")
	| codeFor2 (Or _, s1, s2) =
	   (Str "(") * s1 * (Str ")") * || * (Str "(") * s2 * (Str ")")
	| codeFor2 (Implies _, s1, s2) =
	   (Str "(") * s1 * (Str ")") * (Str " -> ") * (Str "(") * s2 * (Str ")")
	| codeFor2 (Eq _, s1, s2) =
	   (Str "(") * s1 * (Str ")") * (Str " = ") * (Str "(") * s2 * (Str ")")
	| codeFor2 (NotEq _, s1, s2) =
	   (Str "! ((") * s1 * (Str ")") * (Str " = ") * (Str "(") * s2 * (Str "))")
	| codeFor2 (Au _, s1, s2) = (Str "A[") * s1 * (Str "U") * s2 * (Str "]")
	| codeFor2 (Eu _, s1, s2) = (Str "E[") * s1 * (Str "U") * s2 * (Str "]")
	| codeFor2 (Ar _, s1, s2) = codeFor1(Not dp,
					     codeFor2(Eu dp,
						      codeFor1(Not dp, s1),
						      codeFor1(Not dp, s2)))
	| codeFor2 (Er _, s1, s2) = codeFor1(Not dp,
					     codeFor2(Au dp,
						      codeFor1(Not dp, s1),
						      codeFor1(Not dp, s2)))
	| codeFor2 (Object {def = d, ...}, s1, s2) = codeFor2 (d, s1, s2)
	| codeFor2 (ObjectInst {obj = d, ...}, s1, s2) = codeFor2 (d, s1, s2)
	| codeFor2 (Builtin {name=n, ...}, s1, s2) = codeFor2 (n, s1, s2)
	| codeFor2 (x, s1, s2) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-2 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s1)^", "^(Str2string s1)^")")

      (* Translate `formula' (ParseTree) into SMV (Str).  If you even
         encounter a DefVar, translate it with `translateDef'. *)

      fun formulaStr(options, varMap, translateDef) formula =
	  let val funName = "formulaStr"
	      val debug = lazyVerbDebug options funName
	      val _ = pushFunStackLazy(funName, fn()=>(pt2string formula)
				       ^", varMap = "^(map2str varMap))
	      fun formStr a =
		  let val funName = funName^"/formStr"
		      val _ = pushFunStackLazy(funName, fn()=>pt2string a)
		      val res = 
			  (case getMapOpt(varMap, a) of
			       SOME x => x
			     | NONE => 
				   (case a of
					(Appl (_, x, TupleExpr (_, [v1, v2]))) =>
					    codeFor2 (x, formStr v1, formStr v2)
				      | (Appl (_, x, v)) => codeFor1 (x, formStr v)
				      | (Next v) => (Str "next(")*(formStr v)*(Str ")")
				      | (Builtin{name=True _, ...}) => truth
				      | (Builtin{name=False _, ...}) => falsity
				      | (Builtin{name=Undefined _, ...}) => bottom
				      | True2 => Str "1"
				      | False2 => Str "0"
				      | DefVar _ => translateDef a
				      | x => raise SympBug
					    ("TransSMV.formulaStr: unexpected expression: "
					     ^(pt2stringDebug x))))
		      val _ = popFunStackLazy(funName, fn()=>Str2string res)
		  in
		      res
		  end
	      val _ = debug(fn()=>"varMap = ["
			    ^(strlist2str ",\n          "
			      (hash2any(fn(v,v')=>(pt2string v)^" -> "^(Str2string v')) varMap))
			    ^"]\n")
	      val str = formStr formula
	      val _ = popFunStackLazy(funName, fn()=>(Str2string str))
	  in
	      str
	  end

      fun genVarDecl(map, TVmap) v =
	  let val funName = "TransSMV.genVarDecl"
	      val _ = pushFunStackLazy(funName, fn()=>pt2string v)
	      val tvals = getTVMap(TVmap, v)
	      val tvalsStr = Strlist2Str ", " tvals
	      val res = (getMap(map, v))*(Str " : {") *  tvalsStr * (Str "};")
	      val _ = popFunStackLazy(funName, fn()=>Str2string res)
	  in res
	  end

      fun genDefineDecl map (v, def) =
	  let val funName = "TransSMV.genDefineDecl"
	      val _ = pushFunStackLazy(funName, fn()=>(pt2string v))
	      val res = (Str ("-- aux. var "^(pt2string v)^"\n"))
			 *(getMap(map, v))*(Str " := ")*def*(Str ";")
	      val _ = popFunStackLazy(funName, fn()=>Str2string res)
	  in res
	  end

      (* Only add a var it if doesn't exist yet. *)
      fun addNewVar(findObject, abs, map, TVmap, booleanSeen) v =
	  let fun getName True2 = Str "1"
		| getName False2 = Str "0"
		| getName v = (case (getMapOpt (map, v)) of
				   SOME n => n
				 | NONE => Str (newName ()))
	      fun addType(t, SOME vlist) =
	          let val pairs = List.map(fn v=>(getName v, v)) vlist
		      val names = List.map #1 pairs
		      fun addToMap(name, v) =
			  (case (getMapOpt (map, v)) of
			       NONE => (addMapDestructive(map, v, name); ())
			     | SOME _ => ())
		  in
		      (addTVMapDestructive (TVmap, v, bottom::names);
		       List.app addToMap pairs)
		  end
		| addType(t, NONE) =
		  raise ProverError
		      ("Infinite type in translation to SMV:\n  Type: "
		       ^(pt2stringDebug t)
		       ^"\n  Expr: "^(pt2string v))
	      fun makeNew() =
	          let val _ = addMapDestructive (map, v, getName v)
		      val options = getOptions()
		      val limit = get_limit options
		      val t = abstractType options (findObject, abs) (getExprType findObject v)
		      val tvals = getTypeValues(options, findObject) t
		      val _ = (case t of 
				   BoolType _ => booleanSeen := true
				 | _ => ())
		  in addType(t, tvals)
		  end
	  in (case getMapOpt(map,v) of
		  SOME _ => ()
		| NONE => makeNew())
	  end

      (* Translate a primitive-type expr. def-structure's body into an SMV expression *)
      fun genExpr(parms as (options, map)) E =
	  let fun translateDef v = raise SympBug
	            ("TransSMV.genExpr: unmapped DefVar: "^(pt2string v))
	      val formulaStr = formulaStr(options, map, translateDef)
	      val genExpr = genExpr parms
	      fun loop(DefExprBodyAtomic e) = formulaStr e
		| loop(DefSeqList pairs) =
		  let fun doOne(C, E) =
		          let val (c, e) = (genExpr C, genExpr E)
			  in c * (Str " : ") * e * (Str ";")
			  end
		      val conds = List.map doOne pairs
		  in (Str "\ncase\n  ")*(Strlist2Str "\n  " conds)*(Str "\nesac")
		  end
		| loop E = raise SympBug
		  ("TransSMV.genExpr: primitive-type expression contains"
		   ^" compound subexpressions:\n  "^(def2str(DefExprAtomic E)))
	  in loop E
	  end

      (* Translate a primitivi-type def-structure indo an SMV expression.
         Inline the definitions, do not introduce new DEFINE macros or state vars.

	 FIXME: at this point, we do not translate nondeterministic
	 choices in specs.  It could be done when nondet. choices do
	 not include temporal operators, but if they do, we need to
	 figure out how to deal with that properly. *)

      fun genExprExpand(parms as (findObject, abs, map, TVmap, booleanSeen)) D = 
	  let val addNewVar = addNewVar parms
	      val options = getOptions()
	      val defs = getDefs D
	      val E = getBody D
	      (* Hash translations of def'ed vars *)
	      val hash = makeHashDefault(ptEq, pt2string)
	      (* hash mapping def vars to string to replace them with *)
	      val defHash = makeHashDefault(ptEq, pt2string)
	      (* hash of auxiliary state vars that we have to define,
	         mapped to the list of constraints *)
	      val auxVarHash = makeHashDefault(ptEq, pt2string)
	      (* val formulaStr = formulaStr(options, map, fn v => findHash(hash, v)) *)
	      val _ = List.app(fn (v, D)=>(insertHashDestructive(defHash, v, D); ())) defs
	      fun hasTemporal pt = if isTemporal pt then true
				   else List.exists hasTemporal (ptSubtrees pt)
	      fun fetchDef v = 
		  (case findHash(defHash, v) of
		       SOME[E] => E
		     | SOME(lst as _::_) => 
			   (addNewVar v;
			    insertHashDestructive(hash, v, getMap(map, v));
			    (* Delay translation of aux. state var assignments *)
			    insertHashDestructive(auxVarHash, v, List.map descend lst);
			    DefExprBodyAtomic v)
		     | SOME[] => raise SympBug
			   ("TransSMV.genExprExpand: def. var has empty definition:\n  var: "
			    ^(pt2string v)^"\n in def-structure: "^(def2str D))
		     | NONE => raise SympBug
			   ("TransSMV.genExprExpand: def. var has NO definition:\n  var: "
			    ^(pt2string v)^"\n in def-structure: "^(def2str D)))
	      and substDef v =
		  (case findHash(hash, v) of
		       SOME str => str
		     | NONE => let val E = fetchDef v
				   val res = descend E
			       in (insertHashDestructive(hash, v, res); res)
			       end)
	      and descend(DefExprBodyAtomic e) = formulaStr(options, map, substDef) e
		| descend(DefSeqList pairs) = 
		  let fun doOne(C, E) =
		          let val (c, e) = (descend C, descend E)
			  in c * (Str " : ") * e * (Str ";")
			  end
		      val conds = List.map doOne pairs
		  in (Str "\ncase\n  ")*(Strlist2Str "\n  " conds)*(Str "\nesac")
		  end
		| descend E = raise SympBug
		  ("TransSMV.genExprExpand: primitive-type expression contains"
		   ^" compound subexpressions:\n  "^(def2str(DefExprAtomic E))
		   ^"\n in def-structure: "^(def2str D))
	      fun genAuxDecls() = hash2any(fn(n,_)=> genVarDecl(map, TVmap) n) auxVarHash
	      fun genAuxAssts() =
		  let val pairs = hash2any(fn x=>x) auxVarHash
		      fun doPair(v, Es) = 
			  let val lst = List.map(fn E => (getMap(map, v))*(Str " = ")*E) Es
			  in 
			      (Str "(")*(Strlist2Str ") | (" lst)*(Str ")")
			  end
		  in
		      List.map doPair pairs
		  end
	      val spec = descend E
	      val auxDecls = genAuxDecls()
	      val auxAssts = genAuxAssts()
	  in
	      {spec=spec,
	       auxDecls=auxDecls,
	       auxAssts=auxAssts}
	  end

    (* Get all the values of a given type after trying to abstract it. *)
    fun getTypeValuesAbs(options, findObject, abs) tp =
	let val atp = abstractType options (findObject, abs) tp
	    val typeSize = (case getTypeSize(options, findObject) atp of
				s as FiniteSize n => if n <= (get_limit options) then s
						     else InfiniteSize
			      | InfiniteSize => InfiniteSize)
	in
	    case typeSize of
		InfiniteSize => NONE
	      | FiniteSize _ => getTypeValues(options, findObject) atp
	end

      (* Generate DEFINE macro definitions and assignments for auxiliary
         state vars and add them DESTRUCTIVELY to `macros' and `auxvars'
	 hashes.  On the way, update `map' and `TVmap' with the new vars.
	 Returns the unit type. *)
      fun genAux(parms as (options, findObject, limit, abs, map, TVmap))
                (macros, auxvars: (ParseTree, ParseTree list * Str) Hash,
		 booleanSeen, varList) defs =
	  let val funName = "TransSMV.genAux"
	      val _ = pushFunStackLazyStr(funName, fn()=>
					  def2Str(makeDef([defs], DefExprBodyAtomic Fake)))
	      fun op *(x,y) = Conc(x,y)
	      val exprType = getExprType findObject
	      val genExpr = genExpr(options, map)
	      val addNewVar = addNewVar(findObject, abs, map, TVmap, booleanSeen)
	      val normET = normalizeDefET findObject
	      val normER = normalizeDefER findObject
	      val normEC = normalizeDefEC findObject
	      val normEI = normalizeDefEI findObject
	      val normApp = normalizeDefAppl(findObject, abs, SOME varList)
	      val exprDef = exprDef findObject

	      (* Global hash mapping all the new vars and macros to
	         functions: unit -> unit.  Each such a function would
	         do the necessary work to add the var into the
	         system.

		 We have to delay the computation in order to get all
		 the defs into the hash before we start translating
		 the expressions, since defs may come in arbitrary
		 order and some of them maybe undefined when
		 translating others. *)

	      val hash = makeHashDefault(ptEq, pt2string)

	      (* Common translation.  Split (v, E) by the type,
	         combine results with `comb', and translate the
	         primitive-type vars with `final(v, E)' *)

	      fun doCommon(final, comb) (v, E) =
	          let val funName = funName^"/doCommon"
		      val _ = pushFunStackLazyStr(funName, fn()=>(pt2str v)*(Str ", ")*(def2Str E))
		      val doCommon = doCommon(final, comb)
		      fun split tp = 
			  (case abstract options abs tp of
			       SOME a => final(v, E)
			     | NONE => split' tp)
		      and split'(TupleType(_, lst)) =
			  let fun doOne i = doCommon(ExtractTuple(i, v),
						     normET(i, E))
			  in comb(List.map doOne (List.tabulate(List.length lst, fn x=>x)))
			  end
			| split'(t as RecordType(_, lst)) =
			  let fun doOne(RecordField{name=f,...}) =
			            doCommon(ExtractRecord(f, v),
					     normER(f, E))
				| doOne x = raise SympBug
				   ("TransSMV.genAux/doCommon/split: "
				    ^"not a RecordField in RecordType:\n  "
				    ^(pt2string x)^"\n  Type: "^(pt2string t))
			  in comb(List.map doOne lst)
			  end
			| split'(EnumType(_, lst)) =
			  let val V = exprDef v
			      val ind = doCommon(ExtractIndex v, normEI E)
			      fun doOne(c as TypeConstr{Type=FunType _,...}) =
				    SOME(doCommon(ExtractAppl(c, v),
						  normEC(c, E)))
				| doOne _ = NONE
			  in comb(ind::(List.mapPartial doOne lst))
			  end
			| split'(FunType(_, t1, t2)) =
			  let val vals =
			           (case getTypeValuesAbs(options, findObject, abs) t1 of
					SOME lst => lst
				      | NONE => raise ProverError
					  ("Index type of the array is too large or infinite:\n  "
					   ^(pt2string t1)^"\n  in the definition of  "
					   ^(pt2string v)^" = "^(def2str E)))
			      fun doOne d =
				  doCommon(Appl(dp, v, d),
					   normApp(E, DefExprAtomic(DefExprBodyAtomic d)))
			  in comb(List.map doOne vals)
			  end
			(* The type must be atomic now *)
			| split' _ = final(v, E)
			         
		      val tp = exprType v
		      val res = split tp
		      val _ = popFunStackLazy(funName, fn()=>"Returning")
		  in res
		  end
	      (* Collect additional definitions we create on the given ones *)
	      val newDefsRef = ref(makeHashDefault(ptEq, pt2string))
	      fun doDefine(v, E) = 
		  let fun comb _ = ()
		      val funName = funName^"/doDefine"
		      val _ = pushFunStackLazyStr(funName, fn()=>(pt2str v)*(Str ", ")
						  *(def2Str(DefExprAtomic E)))
		      val hash = makeHashDefault(ptEq, pt2string)
		      val debug = lazyVerbDebug options funName
		      fun ff(v, E) = 
		          let val _ = debug(fn()=>"\n"^funName^": adding pair ("
					    ^(pt2string v)^", "^(def2str E)^")\n")
			      val _ = addNewVar v
		          in 
			      (insertHashDestructive(hash, v, getBody E);
			       List.app(fn (v, Es)=>
					(insertHashDestructive(!newDefsRef, v, Es); ()))
			               (getDefs E))
			  end
		      fun doPair(v, E) = (insertHashDestructive(macros, v, genExpr E); ())
		      val _ = doCommon(ff, comb) (v, DefExprAtomic E)
		      val pairs = hash2any(fn x=>x) hash
		      fun resFn() =
			  (List.app doPair pairs;
			   debug(fn()=>funName^": processing "^(pt2string v)^"\n"))
		      val _ = popFunStackLazy(funName, fn()=>"Returning")
		  in resFn
		  end
	      fun doVar v E =
		  let fun foldAnd lst =
		           (Str "(")*(Strlist2Str " & " lst)*(Str ")")
		      fun comb _ = ()
		      val funName = funName^"/doVar"
		      val _ = pushFunStackLazyStr(funName, fn()=>(pt2str v)*(Str ", ")
						  *(def2Str(DefExprAtomic E)))
		      val debug = lazyVerbDebug options funName
		      val hash = makeHashDefault(ptEq, pt2string)
		      fun ff(v, E) = 
		          let val _ = debug(fn()=>"\n"^funName^": adding pair ("
					    ^(pt2string v)^", "^(def2str E)^")\n")
			      val _ = addNewVar v
		          in 
			      (insertHashDestructive(hash, v, fn()=>getBody E);
			       List.app(fn (v, Es)=>
					(insertHashDestructive(!newDefsRef, v, Es); ()))
			               (getDefs E))
			  end
		      fun pair2str(v, E) = (getMap(map, v)) * (Str " = ") * (genExpr(E()))
		      val _ = doCommon(ff, comb) (v, DefExprAtomic E)
		      val pairs = hash2any(fn x=>x) hash
		      val names = List.map #1 pairs
		      val res = (names, fn() => foldAnd(List.map pair2str pairs))
		      val _ = popFunStackLazy(funName, fn()=>"Returning")
		  in
		      res
		  end
	      (* Shovel the definition into the right hash *)
	      fun doOne(v, []) = raise SympBug
		    ("TransSMV.genAux: def'ed macro has an empty definition:\n  "^(pt2string v))
		(* Deterministic: make it a DEFINE macro *)
		| doOne(v, [E]) = 
		    (insertHashDestructive(hash, v, doDefine(v, E)); ())
		(* Otherwise make it a state var *)
		| doOne(v, Es) =
		  let val funName = funName^"/doOne"
		      val debug = lazyVerbDebug options funName
		      val pairs = List.map (doVar v) Es
		      val namesHash = makeHashDefault(ptEq, pt2string)
		      fun addName n = (insertHashDestructive(namesHash, n, ()); ())
		      val _ = List.app (fn(names,_)=>List.app addName names) pairs
		      val names = hash2any(fn (x,_)=>x) namesHash
		      val Fs = List.map #2 pairs
		      fun resFn() =
			  let val str = (Str "(")
			           *(Strlist2Str " | " (List.map (fn x=>x())Fs))*(Str ")")
			      val _ = debug(fn()=>funName^": processing "
					    ^(pt2string v)^": names=["
					    ^(strlistmap2str ", " pt2string names)
					    ^"], str = "^(Str2string str)^"\n")
			  in (insertHashDestructive(auxvars, v, (names, str)); ())
			  end
		  in
		      (insertHashDestructive(hash, v, resFn); ())
		  end
	      (* Process all the definitions, check if we got some
	         new, process them, and so on until no definitions are
	         left. *)
	      fun loop [] = ()
		| loop defs = 
		  let val _ = List.app doOne defs
		      fun filterNew(x as (v, _)) = 
			    (case findHash(hash, v) of
				 SOME _ => NONE
			       | NONE => SOME x)
		      val newDefs = List.mapPartial filterNew (hash2any (fn x=>x) (!newDefsRef))
		      val _ = newDefsRef := makeHashDefault(ptEq, pt2string)
		  in 
		      loop newDefs
		  end
	  in
	      (loop defs;
	       hash2any(fn(_,x)=>x()) hash;
	       popFunStack(funName, "Returning"))
	  end

      (* Translate the top-level assignment def-structure into 3 boolean formulas for
         init, invar, and trans sections.
	 Also, generate auxiliary vars and their assignments, and additional DEFINE macros. *)
      fun genParts(parms as (options, findObject, limit, abs, map, TVmap))
                  (macros, auxvars, booleanSeen, varList) F =
	  let val funName = "TransSMV.genParts"
	      val _ = pushFunStackLazy(funName, fn()=>defasst2str F)
	      val genAux = genAux parms (macros, auxvars, booleanSeen, varList)
	      val genVarDecl = genVarDecl(map, TVmap)
	      val genDefineDecl = genDefineDecl map
	      val genExpr = genExpr(options, map)
	      type Triple = {init: Str, next: Str, norm: Str}
	      fun genOr(ff: Triple -> Str) lst =
		    (Str "(")*(Strlist2Str ") | (" (List.map ff lst))*(Str ")")
	      fun genAnd(ff: Triple -> Str) lst =
		    (Str "(")*(Strlist2Str ") & (" (List.map ff lst))*(Str ")")
	      fun splitBody(DefAsstBodyAtomic{norm=normDef, next=nextDef, init=initDef}) =
		  {init=genExpr initDef,
		   next=genExpr nextDef,
		   norm=genExpr normDef}
		| splitBody(DefAsstSeqList pairs) =
		  let fun doOne(C, A) =
		          let val {init=init, next=next, norm=norm} = splitBody A
			      val Cstr = genExpr C
			      val (prefix, postfix) = (Cstr * (Str " : "), Str ";")
			  in {init = prefix*init*postfix,
			      next = prefix*next*postfix,
			      norm = prefix*norm*postfix}
			  end
		      val conds = List.map doOne pairs
		      fun makeCase(ff: Triple -> Str) =
			  (Str "\ncase\n  ")
			  *(Strlist2Str "\n  " (List.map ff conds))
			  *(Str "\nesac\n")
		  in {init = makeCase(fn{init=s,...}=>s),
		      norm = makeCase(fn{norm=s,...}=>s),
		      next = makeCase(fn{next=s,...}=>s)}
		  end
		| splitBody(DefAsstGuardList pairs) =
		  let fun doOne(C, A) =
		          let val {init=init, next=next, norm=norm} = splitBody A
			      val Cstr = genExpr C
			      fun ff(c, f) = (Str "(") * c * (Str " & ") * f * (Str ")")
			  in {init=ff(Cstr, init),
			      next=ff(Cstr, next),
			      norm=ff(Cstr, norm)}
			  end
		      val parts = List.map doOne pairs
		  in {init=genOr(fn{init=x,...}=>x) parts,
		      next=genOr(fn{next=x,...}=>x) parts,
		      norm=genOr(fn{norm=x,...}=>x) parts}
		  end
		| splitBody(DefAsstAnd lst) = 
		  let val parts = List.map splitBody lst
		  in {init=genAnd(fn{init=x,...}=>x) parts,
		      next=genAnd(fn{next=x,...}=>x) parts,
		      norm=genAnd(fn{norm=x,...}=>x) parts}
		  end
		| splitBody(DefAsstOr lst) = 
		  let val parts = List.map splitBody lst
		  in {init=genOr(fn{init=x,...}=>x) parts,
		      next=genOr(fn{next=x,...}=>x) parts,
		      norm=genOr(fn{norm=x,...}=>x) parts}
		  end
	      fun splitDef(DefAsstAtomic body) = splitBody body
		| splitDef(D as DefAsstDecl _) = (genAux(getAsstDefs D);
						  splitBody(getAsstBody D))
	      (* Now, run the whole thing *)
	      val {init=initBool,
		   norm=invarBool,
		   next=transBool} = splitDef F
	      val auxPairs = (hash2any(fn x=>x) auxvars)
	      fun getNames(_,(names, _)) = names
	      val auxNames = List.foldr(op @) [] (List.map getNames auxPairs)
(* 	      val auxNames = (case auxPairs of
				  (* Every element should have all the primitive names.
				     Grab the first one. *)
				  (_, (names, _))::_ => names
				| [] => [])
 *)
	      fun pair2bool(n, (_, f)) = (Str ("-- aux. var "^(pt2string n)^"\n  ")) * f
	      val res = {auxDecls=List.map genVarDecl auxNames,
			 (* Bool. formula for each var, defines an assignment *)
			 auxBools=List.map pair2bool auxPairs,
			 (* DEFINE declarations `name := def' *)
			 defines=List.map genDefineDecl (hash2any(fn x=>x) macros),
			 (* Bool. formulas for INIT, INVAR, and TRANS parts *)
			 initBool=initBool,
			 invarBool=invarBool,
			 transBool=transBool}
	      val _ = popFunStackLazy(funName, fn()=>"Returning")
	  in
	      res
	  end (* genParts *)
		    
      fun SMVcode(options, findObject, limit, abs) (specList, varList, F) =
	  (* (Str "MODULE main SPEC 0\n", makeMapping(), makeTVMapping()) *)

	  let val funName = "SMVcode"
	      val debug = lazyVerbDebug options "SMVcode"
	      val _ = pushFunStackLazy(funName, fn()=>"specs=["
				       ^(strlistmap2str ", " def2str specList)
				       ^"]\n  vars=["^(strlistmap2str ", " pt2string varList)
				       ^"]\n  trans="^(defasst2str F))
	      (* Incorporate defs from specList *)
	      (* val F = makeAsstDef((getAsstDefs F)::(List.map getDefs specList), getAsstBody F) *)
	      val map = makeMapping ()
	      val TVmap = makeTVMapping ()
	      val macros = makeHashDefault(ptEq, pt2string)
	      val auxvars = makeHashDefault(ptEq, pt2string)
	      val booleanSeen = ref false
	      val genExpr = genExpr(options, map)
	      val genExprExpand = genExprExpand(findObject, abs, map, TVmap, booleanSeen)
	      val genParts = genParts(options, findObject, limit, abs, map, TVmap)
		                     (macros, auxvars, booleanSeen, varList)
	      val addNewVar = addNewVar(findObject, abs, map, TVmap, booleanSeen)
(*	      val genAux = genAux(options, findObject, limit, map, TVmap)
                                 (macros, auxvars, booleanSeen) 
*)
	      val genVarDecl = genVarDecl(map, TVmap)
	      (* Eliminate repeated entries in a ParseTree list; namely varList *)
	      fun eliminateRepeated lst =
		  let val hash = makeHashDefault(ptEq, pt2string)
		      fun hash2list() = (hash2any(fn (x,_)=>x) hash)
		      fun add x = (insertHashDestructive(hash, x, ()); ())
		  in 
		      (List.app add lst; hash2list())
		  end
	      val varList = eliminateRepeated varList
	      fun op * (x,y) = Conc(x,y)
	      val _ = (addMapDestructive (map, True(dp), truth);
		       addMapDestructive (map, False(dp), falsity))
	      val _ = debug(fn()=>"\n-- SMVcode: generating types for vars:\n  ["
			    ^(ptlist2str "," varList)^"]\n")
	      val str = Str "\nMODULE main\n\n"
	      val _ = debug(fn()=>"\n-- SMVcode: generating var mapping...\n")
	      val _ = List.app addNewVar varList
	      fun genCommentMap v = (Str "-- ")*(pt2str v)*(Str " -> ")*(genVarDecl v)
	      val comments = Strlist2Str "\n"
		  ((Str "-- SMV Code Automatically Generated by SyMP\n\n--Variable mappings:\n")
		   ::(List.map genCommentMap varList))
	      val comments = comments
		  *(Str "\n\n--Mapping of all names (including variables)\n")
		  *(Strlist2Str "\n"
		    (List.map(fn(v,str)=>(Str "-- ") * v * (Str " -> ")* str)
		     ((Str "undefined", bottom)::(hash2any(fn (x,y)=>(pt2str x, y)) map))))

	      val _ = debug(fn()=> Str2string comments)
	      val _ = debug(fn()=>"\n-- SMVcode: generating VAR declarations...\n")
	      val str = comments * str
	      val dummyBoolStr = 
		  if (!booleanSeen) then (Str "")
		  else (dummyBool*(Str ": {")*bottom*(Str ", ")
			*truth*(Str", ")*falsity*(Str "};\n  "))
	      val otherVarsStr = List.map genVarDecl varList
	      val str = str * (Strlist2Str "\n  "
			       ((Str ("\n\nVAR\n  "))::dummyBoolStr::otherVarsStr))
	      val {auxDecls=auxDecls,
		   auxBools=auxBools,
		   defines=defines,
		   initBool=initBool,
		   invarBool=invarBool,
		   transBool=transBool} = genParts F
	      (* Auxiliary state var declarations *)
	      val str = (case auxDecls of
			     [] => str
			   | _ => str * (Str "\n\n -- Auxiliary State Var Declarations\nVAR\n  ")
				 *(Strlist2Str "\n  " auxDecls))
              (* Auxiliary state var assignments *)
	      val str = (case auxBools of
			     [] => str
			   | _ => str * (Str "\n\n -- Auxiliary State Var Assignments\nINVAR\n  ")
				 *(Strlist2Str "\n\nINVAR  " auxBools))
	      (* DEFINE declarations *)
	      val str = str * (Str "\n\nDEFINE\n  ") * (Strlist2Str "\n  " defines)
              (* INIT *)
	      val str = str * (Str "\n\nINIT\n  ")*initBool
              (* INVAR *)
	      val str = str * (Str "\n\nINVAR\n  ")*invarBool
              (* TRANS *)
	      val str = str * (Str "\n\nTRANS\n  ")*transBool
              (* SPEC declarations *)
	      val triples = List.map genExprExpand specList
	      fun triple2str{spec=spec, auxDecls=auxDecls, auxAssts=auxAssts} =
		  (case auxDecls of
		       [] => (Str "")
		     | _ => (Str "\n\nVAR -- aux. spec vars\n  ")*(Strlist2Str "\n  " auxDecls))
		  *(case auxAssts of
			[] => (Str "")
		      | _ => (Str "\n\nINVAR -- aux. spec vars\n  ")
			    *(Str "(")*(Strlist2Str ")\n  & (" auxAssts)*(Str ")"))
		  *(Str "\n\nSPEC ")*spec
	      val str = str * (Strlist2Str "\n\n" (List.map triple2str triples))
		  *(Str "\n\n -- End of SMV File\n")
	      val _ = debug(fn()=>"\n-- SMVcode: Translation finished.\n")
	      val _ = popFunStackLazy(funName, fn()=>Str2string str)
	  in
	      (str, map: VarSMVMapping, TVmap: TypeValueMapping)
	  end (* SMVcode *)
    end (* local *)
  end
