;;;; Major mode for SyMP error messages.

(require 'symp-common)

(defvar symp-status-buffer nil
  "The buffer for displaying the global SyMP status.  This variable is
buffer local, and is normally stored in the SyMP server buffer.")
(make-variable-buffer-local 'symp-status-buffer)

(defun symp-status-find-buffer ()
  "Find the status buffer for the current SyMP session, or create one
if not found."
  (let ((old-buffer (current-buffer))
	(status symp-server-status)
	(process symp-server-process)
	(symp-buffer symp-server-process-buffer)
	(buffer nil))
    (if symp-buffer
	;; there is a symp server buffer, go there
	(progn
	  (set-buffer symp-buffer)
	  (if (and symp-status-buffer
		   (bufferp symp-status-buffer)
		   (buffer-live-p symp-status-buffer))
	      ;; the status buffer already exists, just take it
	      (setq buffer symp-status-buffer)
	    ;; no status buffer, create a new one
	    (progn
	      (setq buffer (generate-new-buffer "*SyMP Status*"))
	      ;; register it in the server buffer
	      (setq symp-status-buffer buffer)
	      ;; and set up the status buffer properly
	      (set-buffer buffer)
	      (setq symp-server-process process)
	      (setq symp-server-status status)
	      (setq symp-server-process-buffer symp-buffer)
	      (symp-status-mode)))
	  (set-buffer old-buffer))
      ;; if there is no symp buffer, give up, leave `buffer' nil.
      )
    buffer))

; (defun symp-status-erase-buffer ()
;   "Erases the SyMP status buffer for the current SyMP session and hides
; it if it's displayed."
;   (let ((old-buffer (current-buffer))
; 	(err-buffer (symp-status-find-buffer))
; 	(start-point 0)
; 	(windows nil))
;     (if err-buffer
; 	(progn
; 	  (set-buffer err-buffer)
; 	  (erase-buffer)
; 	  ;; Search for all the windows that display the buffer
; 	  (walk-windows 
; 	   '(lambda (window) 
; 	      (if (equal err-buffer (window-buffer window))
; 		  (add-to-list 'windows window)))
; 	   'no t)
; 	  ;; Delete all the collected windows
; 	  (while windows
; 	    (delete-window (car windows))
; 	    (setq windows (cdr windows)))
; 	  (set-buffer old-buffer)))))

(defun symp-status-show-current (status)
  "Pop up the status buffer and display the STATUS string in it."
  (let ((old-buffer (current-buffer))
	(err-buffer (symp-status-find-buffer))
	(start-point 0))
    (if err-buffer
	(progn
	  ;; put the message into the status buffer
	  (set-buffer err-buffer)
	  (erase-buffer)
	  (goto-char (point-max))
	  (setq start-point (point))
	  (insert status)
	  (goto-char start-point)
	  ;; go back and display the status buffer in another window
	  (set-buffer old-buffer)
	  (display-buffer err-buffer t))
      ;; Ouch, no status buffer!  So do the best we can...
      (message status))))

(defun symp-status-mode ()
  "SyMP major mode for displaying current status."
  (interactive)
  ;;; Define the major mode
  (setq major-mode 'symp-status-mode)
  ;; Watch the status of the inferior SYMP process
  (setq mode-line-process 'symp-server-status)  
  (setq mode-name "SyMP Status")
  ;; Register the buffer with the server
  (symp-server-register-buffer))

(provide 'symp-status-mode)
