signature LIST_OPS =
  sig
    val union : ('a * 'a -> bool) -> 'a list list -> 'a list
    val intersect : ('a * 'a -> bool) -> 'a list list -> 'a list
    val subtract: ('a * 'a -> bool) -> ('a list * 'a list) -> 'a list
    val zip: ('a list) * ('b list) -> ('a * 'b) list
    val zipOpt: ('a list) * ('b list) -> ('a * 'b) list option

    (* Remove redundant elements from the list, making it a "set".
       The order of remaining elements is preserved. *)
    val removeRedundant: ('a * 'a -> bool) -> 'a list -> 'a list

    (* Number the elements of the given list (starting from 0) and
       apply `ff' to every element as ff(i, element_i).  Return the
       list of the results. *)
    val mapEnumerate: (int * 'a -> 'b) -> ('a list) -> 'b list
    (* Given an n-list of lists of values, return a list of
       n-lists of all combinations of those values. *)
    val comblist: 'a list list -> 'a list list

    val listEq: ('a * 'a -> bool) -> ('a list * 'a list) -> bool
    val subset: ('a * 'a -> bool) -> ('a list * 'a list) -> bool
    val eqset: ('a * 'a -> bool) -> ('a list * 'a list) -> bool

    (* Given a list of elements, generate the list of all possible
       equivalence relations over those elements.

       Each equivalence relation is represented as a list of lists of
       equivalent elements. *)
    val genEquivClasses: 'a list -> 'a list list list

    val maxList: ('a * 'a -> bool) * 'a -> 'a list -> 'a
    (* Sort list in ascending order *)
    val sortList: ('a * 'a -> bool) -> 'a list -> 'a list
  end
