structure ListOps:> LIST_OPS =
  struct

    open SympBug

    (* Commonly used list functions *)
    fun zip(h1::l1,h2::l2) = (h1,h2)::(zip(l1,l2))
      | zip([],[]) = []
      | zip(_,_) = raise SympBug("ListOps/zip: lists are not the same size")

    fun zipOpt ([],[]) = SOME []
      | zipOpt (x1::l1, x2::l2) = Option.map(fn l=>(x1,x2)::l)(zipOpt(l1,l2))
      | zipOpt _ = NONE

    (* Number the elements of the given list (starting from 0) and
       apply `ff' to every element as ff(i, element_i).  Return the
       list of the results. *)
 
    fun mapEnumerate ff lst =
	let fun loop(n, []) = []
	      | loop(n, h::t) = (ff(n, h))::(loop(n+1, t))
	in 
	    loop(0, lst)
	end

    fun union _ [] = []
      | union _ [l] = l
      | union eqf (h::t) = 
      let 
	val U = union eqf t
      in
	(List.filter (fn x => not (List.exists (fn y => eqf (x, y)) U)) h) @ U
      end

    (* Set-like operations on lists *)
    fun intersect _ [] = []
      | intersect _ [l] = l
      | intersect eq (h::t) =
	let val U = intersect eq t
	in List.filter(fn x=>List.exists(fn y=>eq(x,y)) h) U
	end

    fun subtract _ ([], _) = []
      | subtract eq (el::l1, l2) =
	  if List.exists(fn x=> eq(el, x)) l2 then
	      subtract eq (l1, l2)
	  else el::(subtract eq (l1, l2))

    (* Remove redundant elements from the list, making it a "set".
       The order of remaining elements is preserved. *)
    fun removeRedundant eq lst =
	let fun loop(new, []) = new
	      | loop(new, x::old) =
	        if List.exists(fn y => eq(x, y)) new then loop(new, old)
		else loop(x::new, old)
	in
	    List.rev(loop([], lst))
	end

    (* Given an n-list of lists of values, return a list of
       n-lists of all combinations of those values. *)
    fun comblist [] = [[]]
      | comblist (hd::tl) =
	let val lst = comblist tl
	    fun ff x = List.map(fn l=>x::l) lst
	in List.foldr(op @) [] (List.map ff hd)
	end

    fun listEq eq (lst1,lst2) =
	(case zipOpt(lst1, lst2) of
	     NONE => false
	   | SOME l => List.all eq l)

    (* Check if l1 is a subset of l2 w.r.t. the equality `eq'.
       Actually, `eq' could be an inequality, yielding the obvious
       extended notion of subset. *)
    fun subset eq (l1,l2) = List.all(fn x=>List.exists (fn y=>eq(x,y)) l2) l1
    fun eqset eq (l1,l2) = subset eq (l1,l2) andalso subset eq (l2,l1)

    (* Given a list of elements, generate the list of all possible
       equivalence relations over those elements.

       Each equivalence relation is represented as a list of lists of
       equivalent elements. *)

    fun genEquivClasses [] = []
      | genEquivClasses [x] = [[[x]]]
      | genEquivClasses (el::lst) = 
	let fun pickOne([], acc) = []
	      | pickOne (h::tl, acc) = (h,tl@acc)::(pickOne(tl, h::acc))
	    fun injectToRel x classes = 
		let val picks = pickOne(classes, [])
		    fun makeRel(class, rest) = (x::class)::rest
		in 
		    ([x]::classes)::(List.map makeRel picks)
		end
	    fun inject x rels = List.foldr(op @) [] (List.map(injectToRel x) rels)
	    fun genPick (el, rest) =
		let val rels = genEquivClasses rest
		in inject el rels
		end
	in
	    genPick (el, lst)
	end

    (* Return the maximum element of the list greater than `min', or `min' otherwise. *)
    fun maxList(op <, min) lst =
	let fun loop(c::lst, acc) =
	    if acc < c then loop(lst, c) else loop(lst, acc)
	      | loop ([], acc) = acc
	in loop(lst, min)
	end

    (* Sort list in ascending order *)
    fun sortList _ [] = []
      | sortList _ [x] = [x]
      | sortList (op <) lst =
	let fun split(l, r, []) = (sortList(op <) l, sortList(op <) r)
	      | split(l, r, [x]) = split(x::l, r, [])
	      | split(l, r, x::y::lst) = split(x::l, y::r, lst)
	    fun merge([], r) = r
	      | merge(l, []) = l
	      | merge(l0 as x::l, r0 as y::r) =
		if y < x then y::(merge(l0, r))
		else x::(merge(l, r0))
	in
	    merge(split([], [], lst))
	end
  end
