structure Str: STR =
  struct
    datatype Str = Str of string | Conc of Str * Str

    (* Concatenate strings "logarithmically": rebuild Str value for
       approx. 128-char strings, then double the threshold and rebuild
       again until we have just one string. *)
    fun Str2string x =
	let fun concat n (x1 as Str s1, x2 as Str s2) =
	        if (String.size s1) + (String.size s2) <= n then Str(s1^s2)
		else Conc(x1, x2)
	      | concat _ (x1, x2) = Conc(x1, x2)
	    fun StrGroup (n: int) (Conc(s1, s2)) = concat n (StrGroup n s1, StrGroup n s2)
	      | StrGroup _ s = s
	    fun loop _ (Str s) = s
	      | loop n s = loop (2*n) (StrGroup n s)
	    (* fun Str2list (Str s) acc = (List.rev (String.explode s))@acc
	      | Str2list (Conc(s1, s2)) acc = Str2list s2 (Str2list s1 acc) *)
	in
	    (* String.implode(List.rev(Str2list x [])) *)
	    loop 128 x
	end

(*      fun Str2string (Conc(s1,s2)) = (Str2string s1)^(Str2string s2) *)
(*        | Str2string (Str s) = s *)

    fun strlist2str sep [] = ""
      | strlist2str sep [hd] = hd
      | strlist2str sep (hd::tl) = hd^sep^(strlist2str sep tl)

    fun Strlist2Str sep [] = (Str "")
      | Strlist2Str sep [hd] = hd
      | Strlist2Str sep (hd::tl) = Conc(Conc(hd, Str sep), Strlist2Str sep tl)

    fun strlistmap2str sep ff lst = strlist2str sep (List.map ff lst)
    fun Strlistmap2Str sep ff lst = Strlist2Str sep (List.map ff lst)

    fun bool2str true = "true"
      | bool2str false = "false"

    (* Some convenient functions for optional arguments *)
    fun option2string str NONE = str
      | option2string _ (SOME s) = s

    fun option2Str str NONE = str
      | option2Str _ (SOME s) = s

    (* Run the function on the value or return the `default' one if
       the value is NONE.
       Useful for converting optional values to strings. *)
    fun optionPrint ff def NONE = def
      | optionPrint ff _ (SOME v) = ff v

  end
